from .mapping import add_stamen_basemap
from .mapping import add_north_arrow
from .mapping import add_scale_bar
from .mapping import ax_ticks
from .mapping import latlon_to_utm_epsg
from .mapping import set_min_axes
from .mapping import set_figure_size
from .mapping import add_POI_marker
from .mapping import stamen_zoom_level
from .epsg_defs import webmerc, wgs84
from .webmap import BindColormap


# -*- coding: utf-8 -*-
"""AshDisperse -- a steady-state volcanic ash dispersion model.

AshDisperse implements an efficient and accurate numerical method for solving
the advection-diffusion-sedimentation equation for tephra classes of different
characteristics in a wind field.

Example:
    params, met = setup(gui=False)

    result = solve(params, met, timer=True)

    for grain_i in range(0, params.grains.bins):
        result.plot_settling_flux_for_grain_class(grain_i)
        result.plot_conc_for_grain_class(grain_i)
        result.plot_iso_conc_for_grain_class(grain_i, 1e-4)

    result.plot_ashload(resolution=500., vmin=1e-4, nodata=-1,
                        export_gtiff=True, export_name='AshLoad.tif')
"""
import os

os.environ["OMP_NUM_THREADS"] = "1"
os.environ["OPENBLAS_NUM_THREADS"] = "1"
os.environ["MKL_NUM_THREADS"] = "1"
os.environ["VECLIB_MAXIMUM_THREADS"] = "1"
os.environ["NUMEXPR_NUM_THREADS"] = "1"

import time

import numpy as np
from scipy.fft import ifft2

from .config import (
    get_max_threads,
    get_num_threads,
    set_default_threads,
    set_num_threads,
)
from .containers import ChebContainer, VelocityContainer
from .core import AshDisperseResult
from .interface import set_met, set_parameters
from .solver import ade_ft_system, source_xy_dimless
from .spectral import grid_freq

set_default_threads(1)


def setup(gui=False):
    """Set parameters and meteorological data for AshDisperse.

    The Parameters class object stores the parameters required for the model
    and the MetData class object stores meteorological data.
    setup() instantiates and initializes these objects.

    Args:
        gui (bool, optional): Use a tkinter gui; defaults to False.

    Returns:
        params_set (Parameters): A Parameters object containing parameters.
        met_set (MetData): A MetData object containing meteorological data.
    """
    params_set = set_parameters()
    params_set, met_set = set_met(params_set, gui=gui)
    return params_set, met_set


def solve(parameters, met_data, timer=False):
    """Run the numerical solver using parameters and meteorological data.

    Args:
        parameters (Parameters): A set of parameters contained in a Parameters
                                 object.
        met_data (MetData): Meteorological data contained in a MetData object.
        timer (bool, optional): Run a timer of stages of the solver;
                                defaults to False.

    Returns:
        AshDisperseResult: An AshDisperseResult object containing the model
                           results.
    """
    cheby = ChebContainer(parameters)

    velocities = VelocityContainer(parameters, met_data, cheby.x)

    x, kx = grid_freq(parameters.solver.Nx)
    y, ky = grid_freq(parameters.solver.Ny)

    _, fxy_f = source_xy_dimless(x, y, parameters)
    # fxy,fxy_f = source_xy_dimless_flat(x,y,params)

    res_x = x[1] - x[0]
    res_y = y[1] - y[0]

    if timer:
        start = time.time()
    conc_0_FT = np.zeros(
        (parameters.solver.Ny, parameters.solver.Nx, parameters.grains.bins),
        dtype=np.complex128,
    )
    conc_z_FT = np.zeros(
        (
            parameters.solver.Ny,
            parameters.solver.Nx,
            parameters.output.Nz,
            parameters.grains.bins,
        ),
        dtype=np.complex128,
    )
    if parameters.solver.rk:
        conc_0_FT, conc_z_FT = ade_ft_system_rk(kx, ky, fxy_f, params, Met)
    else:
        conc_0_FT, conc_z_FT = ade_ft_system(
            kx, ky, fxy_f, cheby, parameters, velocities
        )

    if timer:
        mid = time.time()
    conc_0 = np.zeros(np.shape(conc_0_FT))
    conc_z = np.zeros(np.shape(conc_z_FT))
    for igrain in range(0, parameters.grains.bins):
        conc_0[:, :, igrain] = np.real(ifft2(conc_0_FT[:, :, igrain])) / (res_x * res_y)
        for k in range(0, parameters.output.Nz):
            conc_z[:, :, k, igrain] = np.real(ifft2(conc_z_FT[:, :, k, igrain])) / (
                res_x * res_y
            )
    if timer:
        end = time.time()

    if timer:
        print("Equation solve time : ", mid - start)
        print("Inverse FFT time : ", end - mid)
        print("Total solve time : ", end - start)

    return AshDisperseResult(parameters, x, y, conc_0, conc_z)


if __name__ == "__main__":

    print("Running AshDisperse")

    params = set_parameters()

    Met, params = set_met(params)

    result = solve(params, Met, timer=True)

    for grain_i in range(0, params.grains.bins):
        result.plot_settling_flux_for_grain_class(grain_i)
        result.plot_conc_for_grain_class(grain_i)
        result.plot_iso_conc_for_grain_class(grain_i, 1e-4)

    _ = result.get_ashload(
        resolution=500.0,
        vmin=1e-4,
        nodata=-1,
        export_gtiff=True,
        export_name="AshLoad.tif",
    )

    result.plot_ashload(resolution=500, vmin=0.1)



from .interface import (
    load_inputs,
    save_inputs,
    set_grain_parameters,
    set_met,
    set_output_parameters,
    set_parameters,
    set_physical_parameters,
    set_solver_parameters,
    set_source_parameters,
)



import os
from collections import OrderedDict
from datetime import datetime
from math import ceil, floor

import cdsapi
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import requests
import xarray as xr
from matplotlib import ticker
from matplotlib.colors import Normalize
from netCDF4 import Dataset
from numba import float64
from numba.experimental import jitclass
from numba.types import unicode_type

from ..queryreport import print_text
from ..utilities.utilities import (
    SA_Density_array,
    SA_Density_value,
    SA_Pressure_array,
    SA_Pressure_value,
    SA_Temperature_array,
    SA_Temperature_value,
    interp_ex_array,
    interp_ex_value,
)

MetData_spec = OrderedDict()
MetData_spec["z_data"] = float64[::1]
MetData_spec["temperature_data"] = float64[::1]
MetData_spec["pressure_data"] = float64[::1]
MetData_spec["density_data"] = float64[::1]
MetData_spec["wind_U_data"] = float64[::1]
MetData_spec["wind_V_data"] = float64[::1]
MetData_spec["source"] = unicode_type
MetData_spec["surface_temperature"] = float64  # surface temperature
MetData_spec["surface_pressure"] = float64  # surface pressure
MetData_spec["lapse_tropos"] = float64  # lapse rate in troposphere
MetData_spec["lapse_stratos"] = float64  # lapse rate stratosphere
MetData_spec["height_tropos"] = float64  # height of the troposphere
MetData_spec["height_stratos"] = float64  # height of the stratosphere
MetData_spec["Ra"] = float64  # gas constant of dry air
MetData_spec["g"] = float64  # gravitational acceleration


@jitclass(MetData_spec)
class MetData:
    def __init__(
        self,
        source="standardAtmos",
        surface_temperature=293,
        surface_pressure=101325,
        lapse_tropos=6.5e-3,
        lapse_stratos=2.0e-3,
        height_tropos=11e3,
        height_stratos=20e3,
        Ra=285,
        g=9.81,
    ):
        self.surface_temperature = np.float64(surface_temperature)
        self.surface_pressure = np.float64(surface_pressure)
        self.lapse_tropos = np.float64(lapse_tropos)
        self.lapse_stratos = np.float64(lapse_stratos)
        self.height_tropos = np.float64(height_tropos)
        self.height_stratos = np.float64(height_stratos)
        self.Ra = np.float64(Ra)
        self.g = np.float64(g)
        self.source = source

        self.z_data = np.empty((0), dtype=np.float64)
        self.temperature_data = np.empty((0), dtype=np.float64)
        self.pressure_data = np.empty((0), dtype=np.float64)
        self.density_data = np.empty((0), dtype=np.float64)
        self.wind_U_data = np.empty((0), dtype=np.float64)
        self.wind_V_data = np.empty((0), dtype=np.float64)

    def set_zuv_data(self, z, u, v):
        self.z_data = z
        self.wind_U_data = u
        self.wind_V_data = v

    def temperature_array(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            temperature = interp_ex_array(z, self.z_data, self.temperature_data)
        if self.source == "standardAtmos":
            temperature = SA_Temperature_array(
                z,
                self.surface_temperature,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
            )
        return temperature

    def temperature_value(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            temperature = interp_ex_value(z, self.z_data, self.temperature_data)
        if self.source == "standardAtmos":
            temperature = SA_Temperature_value(
                z,
                self.surface_temperature,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
            )
        return temperature

    def pressure_array(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            pressure = interp_ex_array(z, self.z_data, self.pressure_data)
        elif self.source == "standardAtmos":
            pressure = SA_Pressure_array(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return pressure

    def pressure_value(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            pressure = interp_ex_value(z, self.z_data, self.pressure_data)
        elif self.source == "standardAtmos":
            pressure = SA_Pressure_value(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return pressure

    def density_array(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            density = interp_ex_array(z, self.z_data, self.density_data)
        elif self.source == "standardAtmos":
            density = SA_Density_array(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return density

    def density_value(self, z):
        if self.source in ["netCDF", "GFS", "GFS Archive"]:
            density = interp_ex_value(z, self.z_data, self.density_data)
        elif self.source == "standardAtmos":
            density = SA_Density_value(
                z,
                self.surface_temperature,
                self.surface_pressure,
                self.lapse_tropos,
                self.lapse_stratos,
                self.height_tropos,
                self.height_stratos,
                self.g,
                self.Ra,
            )
        return density

    def wind_U_array(self, z, scale=None):
        u = interp_ex_array(z, self.z_data, self.wind_U_data)
        if scale is None:
            ret = u
        else:
            ret = u / scale
        return ret

    def wind_U_value(self, z, scale=None):
        u = interp_ex_value(z, self.z_data, self.wind_U_data)
        if scale is None:
            ret = u
        else:
            ret = u / scale
        return ret

    def wind_V_array(self, z, scale=None):
        v = interp_ex_array(z, self.z_data, self.wind_V_data)
        if scale is None:
            ret = v
        else:
            ret = v / scale
        return ret

    def wind_V_value(self, z, scale=None):
        v = interp_ex_value(z, self.z_data, self.wind_V_data)
        if scale is None:
            ret = v
        else:
            ret = v / scale
        return ret

    def wind_speed_array(self, z):
        u = self.wind_U_array(z)
        v = self.wind_V_array(z)
        spd = np.sqrt(u * u + v * v)
        return spd

    def wind_speed_value(self, z):
        u = self.wind_U_value(z)
        v = self.wind_V_value(z)
        spd = np.sqrt(u * u + v * v)
        return spd

    def max_wind_speed(self, H, num=100):
        z = np.arange(0.0, H + 1.0, H / (num - 1), dtype=np.float64)
        spd = self.wind_speed_array(z)
        return np.amax(spd)

    def settling_speed_array(self, params, z, scale=None):
        Ws = self.calculate_settling_speed_array(params, z)
        if scale is None:
            ret = Ws
        else:
            ret = Ws / scale
        return ret

    def settling_speed_value(self, params, z, scale=None):
        Ws = self.calculate_settling_speed_value(params, z)
        if scale is None:
            ret = Ws
        else:
            ret = Ws / scale
        return ret

    def settling_speed_for_grain_class_array(self, params, grain_i, z, scale=None):
        Ws = self.calculate_settling_speed_for_grain_class_array(params, grain_i, z)
        if scale is None:
            ret = Ws
        else:
            ret = Ws / scale
        return ret

    def settling_speed_for_grain_class_value(self, params, grain_i, z, scale=None):
        Ws = self.calculate_settling_speed_for_grain_class_value(params, grain_i, z)
        if scale is None:
            ret = Ws
        else:
            ret = Ws / scale
        return ret

    @staticmethod
    def _solve_settling_function(diam, g, rho_p, rho_a, mu, max_iter=50):

        tolerance = 1e-12

        x0 = np.float64(1e-6)
        x1 = np.float64(10000)

        def _settling_func_white(Re, d, g, rho_p, rho_a, mu):
            gp = (rho_p - rho_a) * g / rho_a
            C1 = 0.25
            C2 = 6.0
            Cd = C1 + 24.0 / Re + C2 / (1 + np.sqrt(Re))
            d3 = d**3
            f = Cd * Re * Re - 4.0 / 3.0 * gp * d3 * (rho_a / mu) ** 2
            return f

        fx0 = _settling_func_white(x0, diam, g, rho_p, rho_a, mu)
        fx1 = _settling_func_white(x1, diam, g, rho_p, rho_a, mu)

        if np.abs(fx0) < np.abs(fx1):
            x0, x1 = x1, x0
            fx0, fx1 = fx1, fx0

        x2, fx2 = x0, fx0
        d = x2

        mflag = np.bool(True)
        steps_taken = 0

        while steps_taken < max_iter and np.abs(x1 - x0) > tolerance:
            fx0 = _settling_func_white(x0, diam, g, rho_p, rho_a, mu)
            fx1 = _settling_func_white(x1, diam, g, rho_p, rho_a, mu)
            fx2 = _settling_func_white(x2, diam, g, rho_p, rho_a, mu)

            if fx0 != fx2 and fx1 != fx2:
                L0 = (x0 * fx1 * fx2) / ((fx0 - fx1) * (fx0 - fx2))
                L1 = (x1 * fx0 * fx2) / ((fx1 - fx0) * (fx1 - fx2))
                L2 = (x2 * fx1 * fx0) / ((fx2 - fx0) * (fx2 - fx1))
                new = L0 + L1 + L2
            else:
                new = x1 - fx1 * (x1 - x0) / (fx1 - fx0)

            if (
                (new < 0.25 * (3 * x0 + x1) or new > x1)
                or (mflag and np.abs(new - x1) >= 0.5 * np.abs(x1 - x2))
                or ((not mflag) and np.abs(new - x1) >= 0.5 * np.abs(x2 - d))
                or (mflag and np.abs(x1 - x2) < tolerance)
                or ((not mflag) and np.abs(x2 - d) < tolerance)
            ):
                new = 0.5 * (x0 + x1)
                mflag = np.bool(True)
            else:
                mflag = np.bool(False)

            fnew = _settling_func_white(new, diam, g, rho_p, rho_a, mu)
            d, x2 = x2, x1

            if fx0 * fnew < 0:
                x1 = new
            else:
                x0 = new

            if np.abs(fx0) < np.abs(fx1):
                x0, x1 = x1, x0

            steps_taken += 1

        return x1, steps_taken

    def calculate_settling_speed_array(self, params, z):

        ws = np.empty((z.size, params.grains.bins), dtype=np.float64)

        rho_a = np.empty((z.size, 1), dtype=np.float64)
        rho_a = self.density_array(z)

        for iz, rho_az in enumerate(rho_a):

            for j, (d, rho_p) in enumerate(
                zip(params.grains.diameter, params.grains.density)
            ):

                Re, steps_taken = self._solve_settling_function(
                    d, params.physical.g, rho_p, rho_az, params.physical.mu
                )
                if steps_taken > 50:
                    raise RuntimeError(
                        "In MetData: _solve_settling_function" + " failed to converge"
                    )
                ws[iz, j] = params.physical.mu * Re / rho_az / d

        return ws

    def calculate_settling_speed_value(self, params, z):

        ws = np.empty((params.grains.bins), dtype=np.float64)

        rho_a = self.density_value(z)

        for j, (d, rho_p) in enumerate(
            zip(params.grains.diameter, params.grains.density)
        ):

            Re, steps_taken = self._solve_settling_function(
                d, params.physical.g, rho_p, rho_a, params.physical.mu
            )
            if steps_taken > 50:
                raise RuntimeError(
                    "In MetData: _solve_settling_function" + " failed to converge"
                )
            ws[j] = params.physical.mu * Re / rho_a / d

        return ws

    def calculate_settling_speed_for_grain_class_array(self, params, grain_i, z):

        ws = np.empty((z.size, 1), dtype=np.float64)

        rho_a = np.empty((z.size, 1), dtype=np.float64)
        rho_a = self.density_array(z)

        for iz, rho_az in enumerate(rho_a):

            d = params.grains.diameter[grain_i]
            rho_p = params.grains.density[grain_i]

            Re, steps_taken = self._solve_settling_function(
                d, params.physical.g, rho_p, rho_az, params.physical.mu
            )
            if steps_taken > 50:
                raise RuntimeError(
                    "In MetData: _solve_settling_function" + " failed to converge"
                )
            ws[iz, 1] = params.physical.mu * Re / rho_az / d

        return ws

    def calculate_settling_speed_for_grain_class_value(self, params, grain_i, z):

        rho_a = self.density_value(z)

        d = params.grains.diameter[grain_i]
        rho_p = params.grains.density[grain_i]

        Re, steps_taken = self._solve_settling_function(
            d, params.physical.g, rho_p, rho_a, params.physical.mu
        )
        if steps_taken > 50:
            raise RuntimeError(
                "In MetData: _solve_settling_function" + " failed to converge"
            )
        ws = params.physical.mu * Re / rho_a / d

        return ws


# pylint: disable=E1101
MetData_type = MetData.class_type.instance_type


def save_met(met, file="meteorology.npz"):
    if os.path.exists(file):
        print(
            "WARNING: {outname} ".format(outname=file)
            + "already exists and will be replaced"
        )

    if met.source == "standardAtmos":
        np.savez(
            file,
            source=met.source,
            surface_temperature=met.surface_temperature,
            surface_pressure=met.surface_pressure,
            lapse_tropos=met.lapse_tropos,
            lapse_stratos=met.lapse_stratos,
            height_tropos=met.height_tropos,
            height_stratos=met.height_stratos,
            Ra=met.Ra,
            g=met.g,
            z_data=met.z_data,
            wind_U_data=met.wind_U_data,
            wind_V_data=met.wind_V_data,
        )
    elif met.source in ["netCDF", "GFS"]:
        np.savez(
            file,
            source=met.source,
            Ra=met.Ra,
            g=met.g,
            z_data=met.z_data,
            wind_U_data=met.wind_U_data,
            wind_V_data=met.wind_V_data,
            temperature_data=met.temperature_data,
            pressure_data=met.pressure_data,
            density_data=met.density_data,
        )


def load_met(met_file):

    if not os.path.exists(met_file):
        raise IOError("AshDisperse meteorological file {} not found".format(met_file))

    met = MetData()
    data = np.load(met_file)

    if str(data["source"]) == "standardAtmos":
        met = MetData(
            source="standardAtmos",
            surface_temperature=np.float64(data["surface_temperature"]),
            surface_pressure=np.float64(data["surface_pressure"]),
            lapse_tropos=np.float64(data["lapse_tropos"]),
            lapse_stratos=np.float64(data["lapse_stratos"]),
            height_tropos=np.float64(data["height_tropos"]),
            height_stratos=np.float64(data["height_stratos"]),
            Ra=np.float64(data["Ra"]),
            g=np.float64(data["g"]),
        )
        met.set_zuv_data(data["Z"], data["U"], data["V"])
    elif str(data["source"]) in ["netCDF", "GFS"]:
        met = MetData(
            source="netCDF", Ra=np.float64(data["Ra"]), g=np.float64(data["g"])
        )
        met.z_data = data["z_data"]
        met.wind_U_data = data["wind_U_data"]
        met.wind_V_data = data["wind_V_data"]
        met.temperature_data = data["temperature_data"]
        met.pressure_data = data["pressure_data"]
        met.density_data = data["density_data"]

    return met


def netCDF_to_Met(met, netcdf_data):
    N = len(netcdf_data.altitude)
    met.z_data = np.empty((N), dtype=np.float64)
    met.z_data[:N] = netcdf_data.altitude[:N]
    met.wind_U_data = np.empty((N), dtype=np.float64)
    met.wind_U_data[:N] = netcdf_data.wind_U[:N]
    met.wind_V_data = np.empty((N), dtype=np.float64)
    met.wind_V_data[:N] = netcdf_data.wind_V[:N]
    met.temperature_data = np.empty((N), dtype=np.float64)
    met.temperature_data[:N] = netcdf_data.temperature[:N]
    met.pressure_data = np.empty((N), dtype=np.float64)
    met.pressure_data[:N] = netcdf_data.pressure[:N]
    met.density_data = np.empty((N), dtype=np.float64)
    met.density_data[:N] = netcdf_data.density[:N]
    met.source = "netCDF"
    return met


def gfs_to_Met(met, gfs_data, Ra=287.058):
    z = gfs_data.variables["Geopotential_height_isobaric"]
    temp = gfs_data.variables["Temperature_isobaric"]
    u = gfs_data.variables["u-component_of_wind_isobaric"]
    v = gfs_data.variables["v-component_of_wind_isobaric"]
    pres_name = temp.coordinates.split()[-1]
    pres = gfs_data.variables[pres_name]

    N = gfs_data.variables[pres_name].size

    met.z_data = np.empty((N), dtype=np.float64)
    met.z_data[:N] = np.flipud(z[:].data.flatten())

    met.wind_U_data = np.empty((N), dtype=np.float64)
    met.wind_U_data[:N] = np.flipud(u[:].data.flatten())

    met.wind_V_data = np.empty((N), dtype=np.float64)
    met.wind_V_data[:N] = np.flipud(v[:].data.flatten())

    met.temperature_data = np.empty((N), dtype=np.float64)
    met.temperature_data[:N] = np.flipud(temp[:].data.flatten())

    met.pressure_data = np.empty((N), dtype=np.float64)
    met.pressure_data[:N] = np.flipud(pres[:].data.flatten())

    met.density_data = np.empty((N), dtype=np.float64)
    met.density_data[:N] = met.pressure_data[:N] / Ra / met.temperature_data[:N]

    met.source = "GFS"
    return met


class netCDF:
    def __init__(self, file, lat, lon, Ra=287.058, g=9.80665):
        self.file = file
        self.lat = lat
        self.lon = lon
        self.Ra = Ra
        self.g = g

        self.altitude = np.empty((0), dtype=np.float64)
        self.wind_U = np.empty((0), dtype=np.float64)
        self.wind_V = np.empty((0), dtype=np.float64)
        self.temperature = np.empty((0), dtype=np.float64)
        self.pressure = np.empty((0), dtype=np.float64)
        self.density = np.empty((0), dtype=np.float64)

    def extract(self):
        nc = Dataset(self.file, "r")
        P = nc.variables["level"][:] * 100

        lats = nc.variables["latitude"][:].data
        lons = nc.variables["longitude"][:].data

        if np.amax(lons) > 180.0:
            lons[lons > 180.0] = 180.0 - lons[lons > 180.0]

        lat_i0, _, lon_i0, _ = _near_lat_lon(self.lat, self.lon, lats, lons)

        Z = nc.variables["z"][0, :, lat_i0, lon_i0] / self.g
        self.altitude = np.float64(np.asarray(np.flipud(Z)))

        Uij = _interp_latlon(self.lat, self.lon, nc.variables["u"], lats, lons)
        Vij = _interp_latlon(self.lat, self.lon, nc.variables["v"], lats, lons)
        Tij = _interp_latlon(self.lat, self.lon, nc.variables["t"], lats, lons)

        self.wind_U = np.float64(np.asarray(np.flipud(Uij)))
        self.wind_V = np.float64(np.asarray(np.flipud(Vij)))
        self.temperature = np.float64(np.asarray(np.flipud(Tij)))

        self.pressure = np.float64(np.asarray(np.flipud(P)))
        self.density = np.float64(np.asarray(np.flipud(P / self.Ra / Tij)))

    def download(self, *, datetime):
        if os.path.isfile(self.file):
            print_text("Met file {} exists and will be overwritten".format(self.file))

        cds = cdsapi.Client()
        cds.retrieve(
            "reanalysis-era5-pressure-levels",
            {
                "product_type": "reanalysis",
                "format": "netcdf",
                "variable": [
                    "geopotential",
                    "temperature",
                    "u_component_of_wind",
                    "v_component_of_wind",
                ],
                "pressure_level": [
                    "1",
                    "2",
                    "3",
                    "5",
                    "7",
                    "10",
                    "20",
                    "30",
                    "50",
                    "70",
                    "100",
                    "125",
                    "150",
                    "175",
                    "200",
                    "225",
                    "250",
                    "300",
                    "350",
                    "400",
                    "450",
                    "500",
                    "550",
                    "600",
                    "650",
                    "700",
                    "750",
                    "775",
                    "800",
                    "825",
                    "850",
                    "875",
                    "900",
                    "925",
                    "950",
                    "975",
                    "1000",
                ],
                "year": str(datetime.year),
                "month": "{:02d}".format(datetime.month),
                "day": "{:02d}".format(datetime.day),
                "time": "{:02d}:00".format(datetime.hour),
                "area": [
                    ceil(self.lat + 0.25),
                    floor(self.lon - 0.25),
                    floor(self.lat - 0.25),
                    ceil(self.lon + 0.25),
                ],
            },
            self.file,
        )


class gfs_archive:
    def __init__(self, cycle_datetime, forecast_hr, lat, lon, Ra=287.058, g=9.80665):

        self.lat = lat
        self.lon = lon % 360
        self.Ra = Ra
        self.g = g

        self.forecast_hr = forecast_hr
        self.cycle_datetime = pd.to_datetime(cycle_datetime)

        self.url = self._aws_site()
        self.idx_url = self.url + ".idx"

        grib_exists = self._check_grib()
        idx_exists = self._check_idx()

        if grib_exists and idx_exists:
            self.idx = self._get_idx_as_dataframe()

            self.levels = self.idx.loc[
                (self.idx["level"].str.match("(\d+(?:\.\d+)?) mb"))
            ].level.unique()

    def _aws_site(self):
        return f"https://noaa-gfs-bdp-pds.s3.amazonaws.com/gfs.{self.cycle_datetime:%Y%m%d/%H}/atmos/gfs.t{self.cycle_datetime:%H}z.pgrb2.0p25.f{self.forecast_hr:03d}"

    def _check_grib(self):
        head = requests.head(self.url)
        check_exists = head.ok
        if check_exists:
            check_content = int(head.raw.info()["Content-Length"]) > 1_000_000
            return check_exists and check_content
        else:
            return False

    def _check_idx(self):
        idx_exists = requests.head(self.idx_url).ok
        return idx_exists

    def _get_idx_as_dataframe(self):
        df = pd.read_csv(
            self.idx_url,
            sep=":",
            names=[
                "grib_message",
                "start_byte",
                "reference_time",
                "variable",
                "level",
                "forecast_time",
                "?",
                "??",
                "???",
            ],
        )

        # Format the DataFrame
        df["reference_time"] = pd.to_datetime(df.reference_time, format="d=%Y%m%d%H")
        df["valid_time"] = df["reference_time"] + pd.to_timedelta(
            f"{self.forecast_hr}H"
        )
        df["start_byte"] = df["start_byte"].astype(int)
        df["end_byte"] = df["start_byte"].shift(-1, fill_value="")
        # TODO: Check this works: Assign the ending byte for the last row...
        # TODO: df["end_byte"] = df["start_byte"].shift(-1, fill_value=requests.get(self.grib, stream=True).headers['Content-Length'])
        # TODO: Based on what Karl Schnieder did.
        df["range"] = df.start_byte.astype(str) + "-" + df.end_byte.astype(str)
        df = df.reindex(
            columns=[
                "grib_message",
                "start_byte",
                "end_byte",
                "range",
                "reference_time",
                "valid_time",
                "variable",
                "level",
                "forecast_time",
                "?",
                "??",
                "???",
            ]
        )

        df = df.dropna(how="all", axis=1)
        df = df.fillna("")

        df["search_this"] = (
            df.loc[:, "variable":]
            .astype(str)
            .apply(
                lambda x: ":" + ":".join(x).rstrip(":").replace(":nan:", ":"),
                axis=1,
            )
        )

        # # Attach some attributes
        # df.attrs = dict(
        #     url=self.idx_url,
        #     description="Inventory index file for the GRIB2 file.",
        #     lead_time=self.forecast_hr,
        #     datetime=self.cycle_datetime,
        # )

        return df

    def read_idx(self, searchString=None):
        """
        Inspect the GRIB2 file contents by reading the index file.

        This reads index files created with the wgrib2 utility.

        Parameters
        ----------
        searchString : str
            Filter dataframe by a searchString regular expression.
            Searches for strings in the index file lines, specifically
            the variable, level, and forecast_time columns.
            Execute ``_searchString_help()`` for examples of a good
            searchString.

            .. include:: ../../user_guide/searchString.rst

        Returns
        -------
        A Pandas DataFrame of the index file.
        """

        # Filter DataFrame by searchString
        if searchString not in [None, ":"]:
            logic = self.idx.search_this.str.contains(searchString)
            if logic.sum() == 0:
                print(
                    f"No GRIB messages found. There might be something wrong with {searchString=}"
                )
            df = self.idx.loc[logic]
        return df

    def download_grib(self, searchString, outFile="./gfs_grib_file.grib2"):

        grib_source = self.url

        # Download subsets of the file by byte range with cURL.
        # > Instead of using a single curl command for each row,
        # > group adjacent messages in the same curl command.

        # Find index groupings
        # TODO: Improve this for readability
        # https://stackoverflow.com/a/32199363/2383070
        idx_df = self.read_idx(searchString)
        li = idx_df.index
        inds = (
            [0]
            + [ind for ind, (i, j) in enumerate(zip(li, li[1:]), 1) if j - i > 1]
            + [len(li) + 1]
        )

        curl_groups = [li[i:j] for i, j in zip(inds, inds[1:])]
        curl_ranges = []
        group_dfs = []
        for i, group in enumerate(curl_groups):
            _df = idx_df.loc[group]
            curl_ranges.append(f"{_df.iloc[0].start_byte}-{_df.iloc[-1].end_byte}")
            group_dfs.append(_df)

            for i, (range, _df) in enumerate(zip(curl_ranges, group_dfs)):

                if i == 0:
                    # If we are working on the first item, overwrite the existing file...
                    curl = f"curl -s --range {range} {grib_source} > {outFile}"
                else:
                    # ...all other messages are appended to the subset file.
                    curl = f"curl -s --range {range} {grib_source} >> {outFile}"
                os.system(curl)

    def profiles(self):

        data_P = np.zeros(self.levels.size)
        data_Z = np.zeros(self.levels.size)
        data_T = np.zeros(self.levels.size)
        data_U = np.zeros(self.levels.size)
        data_V = np.zeros(self.levels.size)

        outFile = "./gfs_grib_file.grib2"

        for j, l in enumerate(self.levels):
            data_P[j] = np.float64(l.replace(" mb", "")) * 100

            self.download_grib(f":HGT:{l}", outFile=outFile)

            Z = xr.load_dataset(outFile, engine="cfgrib")
            data_Z[j] = np.float64(
                Z["gh"]
                .interp(latitude=self.lat, longitude=self.lon, method="cubic")
                .values
            )
            self.download_grib(f":TMP:{l}", outFile="./gfs_grib_file.grib2")
            T = xr.load_dataset(outFile, engine="cfgrib")
            data_T[j] = np.float64(
                T["t"]
                .interp(latitude=self.lat, longitude=self.lon, method="cubic")
                .values
            )
            self.download_grib(f":(?:U|V)GRD:{l}", outFile="./gfs_grib_file.grib2")
            UV = xr.load_dataset(outFile, engine="cfgrib")
            uv_interp = UV.interp(latitude=self.lat, longitude=self.lon, method="cubic")
            data_U[j] = np.float64(uv_interp["u"].values)
            data_V[j] = np.float64(uv_interp["v"].values)

        df = pd.DataFrame(
            columns=["altitude", "temperature", "pressure", "wind_U", "wind_V"]
        )
        df["altitude"] = data_Z
        df["temperature"] = data_T
        df["pressure"] = data_P
        df["wind_U"] = data_U
        df["wind_V"] = data_V

        df = df.dropna()
        df = df.sort_values("altitude", ignore_index=True)

        return df


def gfs_archive_to_Met(met, gfs_data, Ra=287.058):
    z = gfs_data.altitude.values
    temp = gfs_data.temperature.values
    u = gfs_data.wind_U.values
    v = gfs_data.wind_V.values
    pres = gfs_data.pressure.values

    N = gfs_data.pressure.size

    met.z_data = np.empty((N), dtype=np.float64)
    met.z_data[:N] = z[:]

    met.wind_U_data = np.empty((N), dtype=np.float64)
    met.wind_U_data[:N] = u[:]

    met.wind_V_data = np.empty((N), dtype=np.float64)
    met.wind_V_data[:N] = v[:]

    met.temperature_data = np.empty((N), dtype=np.float64)
    met.temperature_data[:N] = temp[:]

    met.pressure_data = np.empty((N), dtype=np.float64)
    met.pressure_data[:N] = pres[:]

    met.density_data = np.empty((N), dtype=np.float64)
    met.density_data[:N] = met.pressure_data[:N] / Ra / met.temperature_data[:N]

    met.source = "GFS Archive"
    return met


def _near_lat_lon(target_lat, target_lon, lats, lons):

    lat_i_near = np.abs(lats - target_lat).argmin()
    lon_i_near = np.abs(lons - target_lon).argmin()

    if target_lat in lats:
        lat_i0 = lat_i_near
        lat_i1 = lat_i_near  # won't need this
    else:
        lat_i0 = lat_i_near if lat_i_near < target_lat else lat_i_near + 1
        lat_i1 = lat_i0 - 1
    if target_lon in lons:
        lon_i0 = lon_i_near
        lon_i1 = lon_i_near  # won't need this
    else:
        lon_i0 = lon_i_near if lon_i_near < target_lon else lon_i_near - 1
        lon_i1 = lon_i0 + 1

    return lat_i0, lat_i1, lon_i0, lon_i1


def _interp_latlon(target_lat, target_lon, data, lats, lons):

    lat_i0, lat_i1, lon_i0, lon_i1 = _near_lat_lon(target_lat, target_lon, lats, lons)

    d_lat = target_lat - lats[lat_i0]
    d_lon = target_lon - lons[lon_i0]

    data_00 = data[0, :, lat_i0, lon_i0]
    data_01 = data[0, :, lat_i1, lon_i0]
    data_10 = data[0, :, lat_i0, lon_i1]
    data_11 = data[0, :, lat_i1, lon_i1]

    data_i0 = data_00 + d_lon * (data_10 - data_00)
    data_i1 = data_01 + d_lon * (data_11 - data_01)
    data_ij = data_i0 + d_lat * (data_i1 - data_i0)

    return data_ij


def wind_scale(max_speed):

    max_log10 = np.log10(max_speed)
    major = []
    minor = []
    for j in range(np.int(max_log10) + 1):
        major.append(10**j)
        for k in range(2, 10):
            if k * 10**j > max_speed:
                break
            minor.append(k * 10**j)

    return major, minor


def wind_plot(met, z, show=True, savename=None):

    fig, ax = plt.subplots()

    U = met.wind_U_array(z)
    V = met.wind_V(z)
    speed = met.wind_speed(z)

    max_speed = speed.max()
    major, minor = wind_scale(max_speed)
    major.append(minor[-1])
    minor.pop(-1)

    ax.set_xlim(-max_speed, max_speed)
    ax.set_ylim(-max_speed, max_speed)
    ax.set_aspect("equal")
    ax.axis("off")

    cmap = plt.cm.viridis
    norm = Normalize(vmin=0, vmax=z[-1])
    sm = plt.cm.ScalarMappable(cmap=cmap, norm=Normalize(vmin=0, vmax=z[-1]))

    ax.quiver(
        0,
        -max_speed,
        0,
        2 * max_speed,
        color="darkgray",
        headwidth=6,
        headlength=10,
        scale=1,
        scale_units="xy",
        angles="xy",
    )
    ax.quiver(
        -max_speed / 2,
        0,
        max_speed,
        0,
        color="darkgray",
        headwidth=1,
        headlength=0,
        scale=1,
        scale_units="xy",
        angles="xy",
    )
    ax.text(0, 1.1 * max_speed, "N")

    for m in major:
        ax.add_artist(plt.Circle((0, 0), m, ec="darkgray", lw=0.5, fill=False))
        ax.text(m * np.cos(45 * np.pi / 180), m * np.sin(45 * np.pi / 180), str(m))
    for m in minor:
        ax.add_artist(plt.Circle((0, 0), m, ec="darkgray", lw=0.25, fill=False))

    for (this_z, this_u, this_v) in zip(z, U, V):
        ax.quiver(
            0,
            0,
            this_u,
            this_v,
            color=cmap(norm(this_z)),
            scale=1,
            scale_units="xy",
            angles="xy",
        )

    fmt = ticker.FuncFormatter(lambda z, pos: "{:g}".format(z * 1e-3))
    cbar = fig.colorbar(sm, format=fmt)
    cbar.ax.set_title("Altitude (km)")

    if savename is not None:
        plt.savefig(savename)

    if show:
        plt.show()
    else:
        plt.close()

    return



from collections import OrderedDict

import numpy as np
from numba import complex128, float64, int64
from numba.experimental import jitclass

from ..spectral.cheb import ChebMat

output_spec = OrderedDict()
output_spec["start"] = float64  # start altitude
output_spec["stop"] = float64  # stop altitude
output_spec["step"] = float64  # step altitude
output_spec["altitudes"] = float64[::1]  # Output altitudes
output_spec["Nz"] = int64  # number of output altitudes
# Chebyshev matrices
output_spec["Cheb_lower"] = complex128[:, ::1]  # at lower altitudes
output_spec["Cheb_upper"] = complex128[:, ::1]  # at upper altitudes


@jitclass(output_spec)
class OutputParameters:
    def __init__(self, start=0, stop=40e3, step=1e3):

        self.start = np.float64(start)
        self.stop = np.float(stop)
        self.step = np.float(step)

        self.set_altitudes()

    def set_altitudes(self):
        if self.stop == self.start:
            self.altitudes = np.asarray([0], dtype=np.float64)
        else:
            self.altitudes = np.arange(
                self.start, self.stop + self.step, self.step, dtype=np.float64
            )
        self.Nz = len(self.altitudes)

    def ChebMats(self, N, H):
        z = self.altitudes
        z_lower = z[z <= H]
        z_upper = z[z > H]

        x_lower = 2 * z_lower / H - 1
        x_upper = 1 - 2 * H / z_upper

        self.Cheb_lower = ChebMat(N, x_lower)
        self.Cheb_upper = ChebMat(N, x_upper)

    def describe(self):
        print("Output parameters for AshDisperse")
        print("  Altitudes (m): ")
        for z in self.altitudes:
            print("    ", z)
        print("********************")


# pylint: disable=E1101
OutputParameters_type = OutputParameters.class_type.instance_type



from collections import OrderedDict
from math import ceil

import numpy as np
from numba import float64
from numba.experimental import jitclass

model_spec = OrderedDict()
# Scales and dimensionless parameters as arrays, one for each grain size
model_spec["SettlingScale"] = float64[::1]  # Settling speed
model_spec["Velocity_ratio"] = float64[::1]  # wind speed/settling speed
model_spec["xyScale"] = float64[::1]  # Horizontal scale
model_spec["Lx"] = float64[::1]  # Dimensionless extent in x
model_spec["Ly"] = float64[::1]  # Dimensionless extent in y
model_spec["cScale"] = float64[::1]  # Concentration
model_spec["QScale"] = float64[::1]  # Source term
model_spec["Peclet_number"] = float64  # Peclet number
model_spec["Diffusion_ratio"] = float64  # Kappa_h/Kappa_v
model_spec["sigma_hat"] = float64[::1]  # source radius/xyScale
model_spec["sigma_hat_scale"] = float64[::1]  # sigma_hat scaled to -pi->pi


@jitclass(model_spec)
class ModelParameters:

    # pylint: disable=too-many-instance-attributes
    # These attributes are all needed in ModelParameters

    def __init__(self):
        self.SettlingScale = np.empty((1), dtype=np.float64)
        self.SettlingScale[:] = np.nan
        self.Velocity_ratio = np.empty((1), dtype=np.float64)
        self.Velocity_ratio[:] = np.nan
        self.xyScale = np.empty((1), dtype=np.float64)
        self.xyScale[:] = np.nan
        self.Lx = np.empty((1), dtype=np.float64)
        self.Lx[:] = np.nan
        self.Ly = np.empty((1), dtype=np.float64)
        self.Ly[:] = np.nan
        self.cScale = np.empty((1), dtype=np.float64)
        self.cScale[:] = np.nan
        self.QScale = np.empty((1), dtype=np.float64)
        self.QScale[:] = np.nan
        self.Peclet_number = np.nan
        self.Diffusion_ratio = np.nan
        self.sigma_hat = np.empty((1), dtype=np.float64)
        self.sigma_hat[:] = np.nan
        self.sigma_hat_scale = np.empty((1), dtype=np.float64)
        self.sigma_hat_scale[:] = np.nan

    def from_params(
        self, solver_params, met_params, source_params, grain_params, physical_params
    ):
        N = met_params.Ws_scale.shape[0]
        self.SettlingScale = np.empty((N), dtype=np.float64)
        self.Velocity_ratio = np.empty((N), dtype=np.float64)
        self.xyScale = np.empty((N), dtype=np.float64)
        self.Lx = np.empty((N), dtype=np.float64)
        self.Ly = np.empty((N), dtype=np.float64)
        self.cScale = np.empty((N), dtype=np.float64)
        self.QScale = np.empty((N), dtype=np.float64)
        self.sigma_hat = np.empty((N), dtype=np.float64)
        self.sigma_hat_scale = np.empty((N), dtype=np.float64)

        self.Peclet_number = (
            physical_params.Kappa_h / met_params.U_scale / source_params.PlumeHeight
        )
        self.Diffusion_ratio = physical_params.Kappa_h / physical_params.Kappa_v
        for j in range(N):
            self.SettlingScale[j] = met_params.Ws_scale[j]
            self.Velocity_ratio[j] = met_params.U_scale / self.SettlingScale[j]
            self.xyScale[j] = (
                met_params.U_scale * source_params.PlumeHeight / self.SettlingScale[j]
            )
            self.Lx[j] = (
                ceil(3 * source_params.radius / self.xyScale[j]) * solver_params.domX
            )
            self.Ly[j] = (
                ceil(3 * source_params.radius / self.xyScale[j]) * solver_params.domY
            )
            self.cScale[j] = (
                grain_params.proportion[j]
                * source_params.MER
                / source_params.radius**2
                / self.SettlingScale[j]
            )
            self.QScale[j] = grain_params.proportion[j] * source_params.MER
            self.sigma_hat[j] = source_params.radius / self.xyScale[j]
            self.sigma_hat_scale[j] = (
                np.pi
                * self.sigma_hat[j]
                / np.sqrt(solver_params.domX * solver_params.domY)
            )

    def from_values(
        self,
        SettlingScale,
        Velocity_ratio,
        xyScale,
        Lx,
        Ly,
        cScale,
        QScale,
        Peclet_number,
        Diffusion_ratio,
        sigma_hat,
        sigma_hat_scale,
    ):
        self.SettlingScale = SettlingScale
        self.Velocity_ratio = Velocity_ratio
        self.xyScale = xyScale
        self.Lx = Lx
        self.Ly = Ly
        self.cScale = cScale
        self.QScale = QScale
        self.Peclet_number = Peclet_number
        self.Diffusion_ratio = Diffusion_ratio
        self.sigma_hat = sigma_hat
        self.sigma_hat_scale = sigma_hat_scale

    def describe(self):
        print("Model parameters for AshDisperse")
        print("  Settling speed scale = ", self.SettlingScale)
        print("  Velocity ratio = ", self.Velocity_ratio)
        print("  concentration scale = ", self.cScale)
        print("  x and y scale = ", self.xyScale)
        print("  Lx = ", self.Lx)
        print("  Ly = ", self.Ly)
        print("  source flux scale = ", self.QScale)
        print("  Peclet number = ", self.Peclet_number)
        print("  Diffusion ratio = ", self.Diffusion_ratio)
        print("********************")


# pylint: disable=E1101
ModelParameters_type = ModelParameters.class_type.instance_type



import os
import warnings

import numba as nb
import numpy as np


def get_threading_layer():

    return nb.config.THREADING_LAYER


def set_threading_layer(thread_layer="omp"):

    thread_layers = ["tbb", "omp", "workqueue"]

    if thread_layer not in thread_layers:
        raise ValueError("Invalid thread layer. Expected one of: %s" % thread_layers)

    nb.config.THREADING_LAYER = thread_layer
    return


def get_max_threads():
    return os.cpu_count()


def get_num_threads():
    return nb.get_num_threads()


def set_num_threads(n):
    max_threads = get_max_threads()
    if n > max_threads:
        warnings.warn(
            "Request more threads than available. Setting to maximum recommended.",
            UserWarning,
        )
        nb.set_num_threads(max_threads - 1)
    elif n == max_threads:
        warnings.warn(
            "Setting number of threads equal to the maximum number of threads incurs a performance penalty.",
            UserWarning,
        )
        nb.set_num_threads(n)
    else:
        nb.set_num_threads(n)

    try:
        np.mkl.set_num_threads_local(1)
    except:
        pass
    return


def set_default_threads(n=None):
    if n is None:
        max_threads = get_max_threads()
        set_num_threads(max_threads - 1)
    else:
        set_num_threads(n)
    return



import os
from math import ceil, floor

import folium
import geopandas as gpd
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pyproj as Proj
import rasterio as rio
import rioxarray as rxa
import utm
import xarray as xa
from matplotlib import ticker
from matplotlib.colors import LogNorm, Normalize
from mpl_toolkits.axes_grid1 import make_axes_locatable
from pandas import concat
from rasterio.warp import Resampling, calculate_default_transform, reproject
from rasterio.windows import get_data_window
from scipy.interpolate import RectBivariateSpline, interp1d
from shapely.geometry import Polygon
from skimage.measure import find_contours, marching_cubes

from ashdisperse.params.grain_params import GrainParameters
from ashdisperse.params.met_params import MetParameters
from ashdisperse.params.model_params import ModelParameters
from ashdisperse.params.output_params import OutputParameters
from ashdisperse.params.params import Parameters
from ashdisperse.params.physical_params import PhysicalParameters
from ashdisperse.params.solver_params import SolverParameters
from ashdisperse.params.source_params import SourceParameters

from ..mapping import (
    BindColormap,
    add_north_arrow,
    add_scale_bar,
    add_stamen_basemap,
    ax_ticks,
    latlon_to_utm_epsg,
    set_figure_size,
    set_min_axes,
    stamen_zoom_level,
    webmerc,
    wgs84,
)
from ..utilities import (
    lin_levels,
    log_levels,
    log_steps,
    nice_round_down,
    nice_round_up,
    pad_window,
    plot_rowscols,
)


def _clip_to_window(raster, x, y, vmin=1e-5, pad=1):
    height, width = raster.shape
    raster_tmp = np.copy(raster)
    raster_tmp[raster < vmin] = -1.0
    window = get_data_window(raster_tmp, nodata=-1.0)
    winpad = pad_window(window, height, width, pad=pad)

    row_start = winpad["row_start"]
    row_stop = winpad["row_stop"]
    col_start = winpad["col_start"]
    col_stop = winpad["col_stop"]
    raster_out = raster[row_start:row_stop, col_start:col_stop]
    x_out = x[col_start:col_stop]
    y_out = y[row_start:row_stop]
    return raster_out, x_out, y_out


def _interpolate(raster, x, y, resolution, extent=None, nodata=-1, vmin=1e-6):

    if extent is not None:
        min_x = extent[0]
        min_y = extent[1]
        max_x = extent[2]
        max_y = extent[3]
    else:
        raster_tmp = np.copy(raster)
        raster_tmp[raster < vmin] = nodata
        window = get_data_window(raster_tmp, nodata=nodata)
        min_x = x[window.col_off]
        min_y = y[window.row_off]
        max_x = x[window.col_off + window.width - 1]
        max_y = y[window.row_off + window.height - 1]

    print("resolution = ", resolution)

    left = floor(min_x / resolution) * resolution
    right = ceil(max_x / resolution) * resolution
    bottom = floor(min_y / resolution) * resolution
    top = ceil(max_y / resolution) * resolution

    width = int((right - left) / resolution) + 1
    height = int((top - bottom) / resolution) + 1
    print("Output raster shape is ({h},{w})".format(h=height, w=width))

    x_out = np.linspace(left, right, num=width, endpoint=True)
    y_out = np.linspace(bottom, top, num=height, endpoint=True)

    raster_tmp, x_tmp, y_tmp = _clip_to_window(raster, x, y, vmin=vmin / 10, pad=12)
    raster_tmp[raster_tmp < vmin / 10] = 0.0

    intrp = RectBivariateSpline(y_tmp, x_tmp, raster_tmp)

    raster_out = intrp(y_out, x_out)

    return raster_out, x_out, y_out


class AshDisperseResult:
    def __init__(self, params, x, y, C0, Cz):
        self.grain_classes = params.grains.bins
        self.params = params

        self.utm = utm.from_latlon(
            self.params.source.latitude, self.params.source.longitude
        )

        self.utmepsg = latlon_to_utm_epsg(
            self.params.source.latitude, self.params.source.longitude
        )

        self.x_dimless = x
        self.y_dimless = y
        self.C0_dimless = C0
        self.Cz_dimless = Cz

        self.C0 = np.zeros_like(C0)
        self.Cz = np.zeros_like(Cz)
        self.SettlingFlux = np.zeros_like(C0)

        self.x = np.zeros((len(x), params.grains.bins))
        self.y = np.zeros((len(y), params.grains.bins))

        for j in range(params.grains.bins):
            self.x[:, j] = x * params.model.xyScale[j] * params.model.Lx[j] / np.pi
            self.y[:, j] = y * params.model.xyScale[j] * params.model.Ly[j] / np.pi
            self.C0[:, :, j] = C0[:, :, j] * params.model.cScale[j]
            self.Cz[:, :, :, j] = Cz[:, :, :, j] * params.model.cScale[j]
            self.SettlingFlux[:, :, j] = (
                params.met.Ws_scale[j] * C0[:, :, j] * params.model.cScale[j]
            )

        df = pd.DataFrame(columns=["Name", "Latitude", "Longitude"])
        df.loc[0] = [
            self.params.source.name,
            self.params.source.latitude,
            self.params.source.longitude,
        ]
        self.source_marker = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df.Longitude, df.Latitude),
            crs=wgs84["init"],
        )

    @classmethod
    def from_netcdf(r, filename):
        "Initialize results from a netcdf file"
        da = xa.load_dataset(filename)

        p = Parameters()
        p.solver = SolverParameters(
            domX=da.domX,
            domY=da.domY,
            minN_log2=da.minN_log2,
            maxN_log2=da.maxN_log2,
            Nx_log2=da.Nx_log2,
            Ny_log2=da.Ny_log2,
            epsilon=da.epsilon,
        )
        p.grains = GrainParameters()
        # TODO: do this properly
        if isinstance(da.diameter, np.float64):
            p.grains.fromvalue(da.diameter, da.density, da.proportion)
        else:
            p.grains.fromlists(da.diameter, da.density, da.proportion)

        p.source = SourceParameters(
            da.latitude,
            da.longitude,
            da.utmcode,
            radius=da.radius,
            PlumeHeight=da.PlumeHeight,
            MER=da.MER,
            duration=da.duration,
            Suzuki_k=da.Suzuki_k,
            Suzuki_peak=da.Suzuki_peak,
            name=da.name,
        )

        p.physical = PhysicalParameters(
            Kappa_h=da.Kappa_h,
            Kappa_v=da.Kappa_v,
            g=da.g,
            mu=da.mu,
        )

        p.met = MetParameters(da.U_scale, np.atleast_1d(da.Ws_scale))

        p.model = ModelParameters()
        p.model.from_params(p.solver, p.met, p.source, p.grains, p.physical)

        p.output = OutputParameters(start=da.start, stop=da.stop, step=da.step)

        r = AshDisperseResult(p, da.x.values, da.y.values, da.C0.values, da.Cz.values)

        return r

    def to_netcdf(self, filename="AshDisperse.nc", compress=True, complevel=5):
        da = xa.Dataset(
            data_vars=dict(
                C0=(["x", "y", "grains"], self.C0_dimless),
                Cz=(["x", "y", "z", "grains"], self.Cz_dimless),
            ),
            coords=dict(
                grains=np.arange(self.params.grains.bins),
                x=self.x_dimless,
                y=self.y_dimless,
                z=self.params.output.altitudes,
            ),
            attrs=dict(
                description="AshDisperse results",
                domX=self.params.solver.domX,
                domY=self.params.solver.domY,
                minN_log2=self.params.solver.minN_log2,
                minN=self.params.solver.minN,
                maxN_log2=self.params.solver.maxN_log2,
                maxN=self.params.solver.maxN,
                Nx_log2=self.params.solver.Nx_log2,
                Nx=self.params.solver.Nx,
                Ny_log2=self.params.solver.Ny_log2,
                Ny=self.params.solver.Ny,
                chebIts=self.params.solver.chebIts,
                epsilon=self.params.solver.epsilon,
                meps=self.params.solver.meps,
                bins=self.params.grains.bins,
                diameter=list(self.params.grains.diameter),
                density=list(self.params.grains.density),
                proportion=list(self.params.grains.proportion),
                name=self.params.source.name,
                latitude=self.params.source.latitude,
                longitude=self.params.source.longitude,
                utmcode=self.params.source.utmcode,
                radius=self.params.source.radius,
                PlumeHeight=self.params.source.PlumeHeight,
                MER=self.params.source.MER,
                duration=self.params.source.duration,
                Suzuki_k=self.params.source.Suzuki_k,
                Suzuki_peak=self.params.source.Suzuki_peak,
                Kappa_h=self.params.physical.Kappa_h,
                Kappa_v=self.params.physical.Kappa_v,
                g=self.params.physical.g,
                mu=self.params.physical.mu,
                U_scale=self.params.met.U_scale,
                Ws_scale=list(self.params.met.Ws_scale),
                SettlingScale=self.params.model.SettlingScale,
                Velocity_ratio=self.params.model.Velocity_ratio,
                xyScale=self.params.model.xyScale,
                Lx=self.params.model.Lx,
                Ly=self.params.model.Ly,
                cScale=self.params.model.cScale,
                QScale=self.params.model.QScale,
                Peclet_number=self.params.model.Peclet_number,
                Diffusion_ratio=self.params.model.Diffusion_ratio,
                sigma_hat=self.params.model.sigma_hat,
                sigma_hat_scale=self.params.model.sigma_hat_scale,
                start=self.params.output.start,
                stop=self.params.output.stop,
                step=self.params.output.step,
                altitudes=self.params.output.altitudes,
                Nz=self.params.output.Nz,
            ),
        )

        if compress:
            # comp = dict(zlib=True, complevel=complevel)
            comp = {"compression": "gzip", "compression_opts": complevel}
            encoding = {var: comp for var in da.data_vars}
        else:
            encoding = {}

        da.to_netcdf(
            path=filename,
            mode="w",
            format="NETCDF4",
            engine="h5netcdf",
            encoding=encoding,
        )

        return

    def _check_valid_grain(self, grain_i):
        if grain_i > self.params.grains.bins:
            raise ValueError(
                "Requested grain class not found; given {}".format(grain_i)
                + " in results with"
                + " {} grain classes.".format(self.params.grains.bins)
            )
        if grain_i < 0:
            raise ValueError("Grain class index must be positive")

    def get_settlingflux_for_grain_class(self, grain_i, vmin=1e-7, masked=False):
        self._check_valid_grain(grain_i)

        x = self.x[:, grain_i]
        y = self.y[:, grain_i]

        flux = self.SettlingFlux[:, :, grain_i]

        flux, x, y = _clip_to_window(flux, x, y, vmin=vmin)

        if flux.size == 0:
            return None

        if np.nanmax(flux) < vmin:
            return None

        if masked:
            flux = np.ma.masked_less(flux, vmin)

        return x, y, flux

    def get_ashload_for_grain_class(self, grain_i, vmin=1e-5, masked=False):
        self._check_valid_grain(grain_i)

        x = self.x[:, grain_i]
        y = self.y[:, grain_i]

        ashload = self.SettlingFlux[:, :, grain_i] * self.params.source.duration

        ashload, x, y = _clip_to_window(ashload, x, y, vmin=vmin)

        if ashload.size == 0:
            return None

        if np.nanmax(ashload) < vmin:
            return None

        if masked:
            ashload = np.ma.masked_less(ashload, vmin)

        return x, y, ashload

    def POI_ashload_for_grain_class(
        self, grain_i: int, lat: float, lon: float, vmin: float = 1e-3
    ):
        self._check_valid_grain(grain_i)

        df = pd.DataFrame(columns=["Name", "Latitude", "Longitude"])
        df.loc[0] = ["POI", lat, lon]
        POI = gpd.GeoDataFrame(
            df,
            geometry=gpd.points_from_xy(df.Longitude, df.Latitude),
            crs=wgs84["init"],
        )

        POI = POI.to_crs(self.utmepsg)

        try:
            ashload = self.raster_ashload_for_grain_class(
                grain_i, vmin=vmin, nodata=np.nan, masked=False, crs=None
            )

            POI_ashload = ashload.interp(
                x=POI.geometry.x, y=POI.geometry.y, method="cubic"
            )

            load = POI_ashload["ash_load"].values[0][0]

            if np.isnan(load):
                return 0
            elif load < 0:
                return 0
            else:
                return load
        except:
            return 0

    def POI_ashload(self, lat: float, lon: float, vmin: float = 1e-3):

        loads = []
        total_load = 0
        for j in range(self.grain_classes):
            this_ashload = self.POI_ashload_for_grain_class(j, lat, lon, vmin)
            loads.append(
                {
                    "diameter": self.params.grains.diameter[j],
                    "density": self.params.grains.density[j],
                    "proportion": self.params.grains.proportion[j],
                    "load": this_ashload,
                }
            )
            total_load += this_ashload

        return total_load, loads

    def max_ashload_for_grain_class(self, grain_i):
        self._check_valid_grain(grain_i)
        return np.nanmax(self.SettlingFlux[:, :, grain_i]) * self.params.source.duration

    def raster_settlingflux_for_grain_class(
        self, grain_i, vmin=1e-6, nodata=np.nan, masked=False, crs=None
    ):

        x, y, ashload = self.get_settlingflux_for_grain_class(
            grain_i, vmin=vmin, masked=masked
        )

        ds = xa.Dataset()
        ds.coords["x"] = x + self.utm[0]
        ds.coords["y"] = y + self.utm[1]

        ds["settling_flux"] = (("y", "x"), ashload)
        ds["settling_flux"].attrs["short_name"] = "settling_flux_{}".format(grain_i)
        ds["settling_flux"].attrs[
            "long_name"
        ] = "settling flux at ground level for grain size {} m".format(
            self.params.grains.diameter[grain_i]
        )
        ds["settling_flux"].attrs["units"] = "kg/m**2/s"
        ds["settling_flux"].rio.write_nodata(nodata, inplace=True)

        ds.x.attrs["units"] = "metres"
        ds.y.attrs["units"] = "metres"

        ds = ds.rio.write_crs(self.utmepsg)

        if crs is not None:
            ds = ds.rio.reproject(crs, resampling=Resampling.cubic)

        return ds

    def raster_ashload_for_grain_class(
        self, grain_i, vmin=1e-2, nodata=np.nan, masked=False, crs=None
    ):

        x, y, ashload = self.get_ashload_for_grain_class(
            grain_i, vmin=vmin, masked=masked
        )

        ds = xa.Dataset()
        ds.coords["x"] = x + self.utm[0]
        ds.coords["y"] = y + self.utm[1]

        ds["ash_load"] = (("y", "x"), ashload)
        ds["ash_load"].attrs["short_name"] = "ash_load_{}".format(grain_i)
        ds["ash_load"].attrs[
            "long_name"
        ] = "ash load at ground level for grain size {} m".format(
            self.params.grains.diameter[grain_i]
        )
        ds["ash_load"].attrs["units"] = "kg/m**2"
        ds["ash_load"].rio.write_nodata(nodata, inplace=True)

        ds.x.attrs["units"] = "metres"
        ds.y.attrs["units"] = "metres"

        ds = ds.rio.write_crs(self.utmepsg)

        if crs is not None:
            ds = ds.rio.reproject(crs, resampling=Resampling.cubic)

        return ds

    def _raster_contour(self, raster, name, cntrs):

        data = raster.data
        x = raster.x.data
        y = raster.y.data
        crs = raster.rio.crs

        data_min = np.nanmin(data)
        data_max = np.nanmax(data)
        data_max = nice_round_up(data_max, mag=10 ** np.floor(np.log10(data_max)))

        fx = interp1d(np.arange(0, len(x)), x)
        fy = interp1d(np.arange(0, len(y)), y)

        cntrs = cntrs[cntrs > data_min]
        cntrs = cntrs[cntrs < data_max]

        for kk, this_cntr in enumerate(cntrs):

            C = find_contours(data, this_cntr)

            if len(C) == 0:
                pass
            for jj, p in enumerate(C):
                p[:, 0] = fy(p[:, 0])
                p[:, 1] = fx(p[:, 1])

                p[:, [0, 1]] = p[:, [1, 0]]

                if len(p[:, 1]) > 2:
                    thisPoly = Polygon(p).buffer(0)
                    if not thisPoly.is_empty:
                        if jj == 0:
                            geom = thisPoly
                            g_tmp = gpd.GeoDataFrame(
                                columns=["contour", "name", "geometry"], crs=crs
                            )
                            g_tmp.loc[0, "contour"] = this_cntr
                            g_tmp.loc[[0], "geometry"] = gpd.GeoSeries([geom])
                            g_tmp.loc[0, "name"] = name
                        else:
                            if g_tmp.loc[0, "geometry"].contains(thisPoly):
                                geom = g_tmp.loc[0, "geometry"].difference(thisPoly)
                            else:
                                try:
                                    geom = g_tmp.loc[0, "geometry"].union(thisPoly)
                                except:
                                    print(
                                        "Error processing polygon contour -- skipping.  Better check the result!"
                                    )
                            g_tmp.loc[[0], "geometry"] = gpd.GeoSeries([geom])

            if kk == 0:
                g = gpd.GeoDataFrame(g_tmp).set_geometry("geometry")
            else:
                g = gpd.GeoDataFrame(pd.concat([g, g_tmp], ignore_index=True))

        g["contour"] = g["contour"].astype("float64")
        g.crs = crs

        return g

    def write_gtiff(self, raster, x, y, outname, nodata=-1, vmin=1e-6, resolution=None):

        print("Writing data to {outname}".format(outname=outname))
        if os.path.isfile(outname):
            print(
                "WARNING: {outname} ".format(outname=outname)
                + "already exists and will be replaced"
            )

        utmcode = self.params.source.utmcode

        if resolution is not None:
            raster, x, y = _interpolate(
                raster, x, y, resolution, nodata=nodata, vmin=vmin
            )

        height, width = raster.shape
        raster[raster < vmin] = nodata

        transform = rio.transform.from_bounds(
            x[0] + self.utm[0],
            y[0] + self.utm[1],
            x[-1] + self.utm[0],
            y[-1] + self.utm[1],
            width,
            height,
        )

        with rio.open(
            outname,
            "w",
            driver="GTiff",
            height=height,
            width=width,
            count=1,
            dtype=str(raster.dtype),
            crs=utmcode,
            nodata=nodata,
            transform=transform,
            resampling=Resampling.cubic,
            compress="lzw",
        ) as dst:
            dst.write(np.flipud(raster), 1)

    def write_settling_flux_for_grain_class(
        self, grain_i, nodata=-1, vmin=1e-6, resolution=None
    ):
        self._check_valid_grain(grain_i)
        self.write_gtiff(
            self.SettlingFlux[:, :, grain_i],
            self.x[:, grain_i],
            self.y[:, grain_i],
            "SettlingFlux_{}.tif".format(grain_i),
            nodata=nodata,
            vmin=vmin,
            resolution=resolution,
        )

    def write_ashload_for_grain_class(
        self, grain_i, vmin=1e-3, resolution=None, outname=None
    ):
        self._check_valid_grain(grain_i)

        ashload = self.raster_ashload_for_grain_class(grain_i, vmin=vmin, masked=False)

        res_x, res_y = ashload.rio.resolution()
        height = ashload.rio.height
        width = ashload.rio.width
        if resolution is not None:
            new_width = round(width * res_x / resolution)
            new_height = round(height * res_y / resolution)
            ashload = ashload.rio.reproject(
                ashload.rio.crs,
                shape=(new_height, new_width),
                resampling=Resampling.cubic,
            )

        if outname is None:
            outname = "AshLoad_{}.tif".format(grain_i)

        print("Writing data to {outname}".format(outname=outname))
        if os.path.isfile(outname):
            print(
                "WARNING: {outname} ".format(outname=outname)
                + "already exists and will be replaced"
            )

        ashload.rio.to_raster(outname)
        return

    def get_ashload(
        self,
        resolution=300.0,
        vmin=1e-3,
        nodata=-1,
        export_gtiff=True,
        export_name="AshLoad.tif",
    ):

        ash_load_grains = self.SettlingFlux * self.params.source.duration
        Ngrains = self.params.grains.bins

        row_off = []
        col_off = []
        h = []
        w = []
        x_min = []
        y_min = []
        x_max = []
        y_max = []

        for igrain in range(Ngrains):
            thisAshLoad = ash_load_grains[:, :, igrain]
            thisAshLoad[thisAshLoad < vmin / Ngrains / 10] = nodata
            window = get_data_window(thisAshLoad, nodata=nodata)
            row_off.append(window.row_off)
            col_off.append(window.col_off)
            h.append(window.height)
            w.append(window.width)
            x_min.append(self.x[window.col_off, igrain])
            x_max.append(self.x[window.col_off + window.width - 1, igrain])
            y_min.append(self.y[window.row_off, igrain])
            y_max.append(self.y[window.row_off + window.height - 1, igrain])

        min_x = min(x_min)
        min_y = min(y_min)
        max_x = max(x_max)
        max_y = max(y_max)

        left = floor(min_x / resolution) * resolution
        right = ceil(max_x / resolution) * resolution
        bottom = floor(min_y / resolution) * resolution
        top = ceil(max_y / resolution) * resolution

        width = int((right - left) / resolution) + 1
        height = int((top - bottom) / resolution) + 1
        print("Output raster shape is ({h},{w})".format(h=height, w=width))

        x_out = np.linspace(left, right, num=width, endpoint=True)
        y_out = np.linspace(bottom, top, num=height, endpoint=True)

        totalAshLoad = np.zeros((height, width), dtype=np.float64)

        for igrain in range(Ngrains):
            thisAshLoad = ash_load_grains[
                row_off[igrain] : row_off[igrain] + h[igrain] - 1,
                col_off[igrain] : col_off[igrain] + w[igrain] - 1,
                igrain,
            ]
            thisAshLoad[thisAshLoad < vmin / Ngrains / 10] = 0
            x = self.x[col_off[igrain] : col_off[igrain] + w[igrain] - 1, igrain]
            y = self.y[row_off[igrain] : row_off[igrain] + h[igrain] - 1, igrain]
            intrp = RectBivariateSpline(y, x, thisAshLoad)

            totalAshLoad += intrp(y_out, x_out)

        totalAshLoad[totalAshLoad < vmin] = nodata
        window = get_data_window(totalAshLoad, nodata=nodata)
        winpad = pad_window(window, height, width)
        totalAshLoad = totalAshLoad[
            winpad["row_start"] : winpad["row_stop"],
            winpad["col_start"] : winpad["col_stop"],
        ]
        x_out = x_out[winpad["col_start"] : winpad["col_stop"]]
        y_out = y_out[winpad["row_start"] : winpad["row_stop"]]

        if export_gtiff:
            self.write_gtiff(
                totalAshLoad,
                x_out,
                y_out,
                export_name,
                nodata=nodata,
                vmin=vmin,
                resolution=None,
            )
        totalAshLoad = np.ma.masked_where(totalAshLoad < vmin, totalAshLoad)
        return x_out, y_out, totalAshLoad

    def contour_settling_flux(self, grain_i, logscale=True, vmin=1e-6):

        self._check_valid_grain(grain_i)
        data = self.SettlingFlux[:, :, grain_i]
        vmax = np.nanmax(data)
        mag = np.log10(vmax)
        if mag > 0:
            mag = ceil(mag)
        else:
            mag = floor(mag)
        vmax = nice_round_up(vmax, mag=10**mag)

        fx = interp1d(np.arange(0, len(self.x)), self.x[:, grain_i] + self.utm[0])
        fy = interp1d(np.arange(0, len(self.y)), self.y[:, grain_i] + self.utm[1])

        if logscale:
            levels = log_levels(vmin, vmax)
        else:
            levels = lin_levels(vmin, vmax)

        g = None
        for level in levels:
            C = find_contours(data, level)

            if len(C) > 0:
                for jj, p in enumerate(C):
                    pxy = np.zeros_like(p)
                    px = fx(p[:, 1])
                    py = fy(p[:, 0])
                    pxy[:, 0] = px
                    pxy[:, 1] = py

                    this_poly = Polygon(pxy)
                    if not this_poly.is_empty:
                        if jj == 0:
                            geom = this_poly
                            g_tmp = gpd.GeoDataFrame(
                                columns=["SettlingFlux", "GrainSize", "geometry"],
                                crs=self.params.source.utmcode,
                            )
                            g_tmp.loc[0, "GrainClass"] = grain_i
                            g_tmp.loc[0, "GrainSize"] = self.params.grains.diameter[
                                grain_i
                            ]
                            g_tmp.loc[0, "SettlingFlux"] = level
                            g_tmp.loc[0, "geometry"] = geom
                        else:
                            if g_tmp.loc[0, "geometry"].contains(this_poly):
                                geom = g_tmp.loc[0, "geometry"].difference(this_poly)
                            else:
                                geom = g_tmp.loc[0, "geometry"].union(this_poly)
                            g_tmp.loc[[0], "geometry"] = gpd.GeoSeries([geom])

            if g is None:
                g = gpd.GeoDataFrame(g_tmp).set_geometry("geometry")
            else:
                g = gpd.GeoDataFrame(concat([g, g_tmp], ignore_index=True))

        g["SettlingFlux"] = g["SettlingFlux"].astype("float64")
        g["GrainClass"] = g["GrainClass"].astype("int64")
        return g

    def plot_settling_flux_for_grain_class(
        self,
        grain_i,
        logscale=True,
        vmin=1e-6,
        vmax=None,
        cmap=plt.cm.Purples,
        basemap=True,
        alpha=0.5,
        max_zoom=None,
        show=True,
        save=False,
        save_name="ashdisperse_result.png",
        min_ax_width=None,
        min_ax_height=None,
    ):
        self._check_valid_grain(grain_i)

        ds = self.raster_settlingflux_for_grain_class(
            grain_i, vmin=vmin, crs=webmerc["init"]
        )
        if ds is None:
            raise RuntimeError("No data in plot_settling_flux_for_grain_class()")

        x = ds.x
        y = ds.y
        data = ds["settling_flux"].values

        x_intrp = interp1d(np.arange(len(x)), x)
        y_intrp = interp1d(np.arange(len(y)), y)

        if vmax is None:
            maxflux = np.nanmax(data)
            mag = np.log10(maxflux)
            if mag > 0:
                mag = ceil(mag)
            else:
                mag = floor(mag)
            vmax = nice_round_up(maxflux, mag=10**mag)
        else:
            mag = np.log10(vmax)
            if mag > 0:
                mag = ceil(mag)
            else:
                mag = floor(mag)

        if vmin is None:
            vmin = 10 ** (mag - 3)

        cbar_fig, cbar_ax = plt.subplots()
        if logscale:
            levels = log_levels(vmin, vmax)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=LogNorm()
            )
        else:
            levels = lin_levels(vmin, vmax, num=20)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=Normalize()
            )

        fig, ax = plt.subplots()
        for j, l in enumerate(levels):
            cntrs = find_contours(data, l)
            for c in cntrs:
                ax.fill(
                    x_intrp(c[:, 1]),
                    y_intrp(c[:, 0]),
                    color=cmap(j / len(levels)),
                    alpha=alpha,
                    zorder=1,
                )

        source = self.source_marker.to_crs(webmerc["init"])
        source.plot(ax=ax, marker="^", color="k", markersize=20, zorder=2)

        xlim = list(ax.get_xlim())
        ylim = list(ax.get_ylim())

        x_width = xlim[1] - xlim[0]
        y_height = ylim[1] - ylim[0]

        min_ax_width = int(1.5 * x_width)
        min_ax_height = int(1.5 * y_height)

        set_min_axes(ax, min_width=min_ax_width, min_height=min_ax_height)

        fig = set_figure_size(fig, ax)

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)

        cbar = plt.colorbar(tmp, ax=ax, cax=cax)
        cbar.minorticks_off()
        cbar.set_label("Settling flux (kg/m\u00B2/s)")
        plt.close(cbar_fig)

        ax = ax_ticks(ax, source.geometry[0].x, source.geometry[0].y)
        if basemap:
            ax = add_stamen_basemap(ax, stamen="background", zorder=0)
            (Narrow, ntext) = add_north_arrow(ax, zorder=11, fontsize=16)
            (scalebar, sbframe) = add_scale_bar(ax, segments=1)

        plt.draw()
        if save:
            plt.savefig(save_name, dpi=300, transparent=True, format="png")

        if show:
            plt.show()
        else:
            plt.close(fig)
        return (fig, ax, cbar)

    def plot_ashload_for_grain_class(
        self,
        grain_i,
        logscale=True,
        vmin=None,
        vmax=10,
        cmap=plt.cm.plasma,
        basemap=True,
        alpha=0.5,
        max_zoom=None,
        show=True,
        save=False,
        save_name="ashdisperse_result.png",
        min_ax_width=None,
        min_ax_height=None,
    ):
        self._check_valid_grain(grain_i)

        ds = self.raster_ashload_for_grain_class(
            grain_i, vmin=1e-6, crs=webmerc["init"]
        )
        if ds is None:
            raise RuntimeError("No data in plot_ashload_for_grain_class()")

        x = ds.x
        y = ds.y
        data = ds["ash_load"].values

        x_intrp = interp1d(np.arange(len(x)), x)
        y_intrp = interp1d(np.arange(len(y)), y)

        if vmax is None:
            maxload = np.nanmax(data)
            mag = np.log10(maxload)
            if mag > 0:
                mag = ceil(mag)
            else:
                mag = floor(mag)
            vmax = nice_round_up(maxload, mag=10**mag)
        else:
            mag = np.log10(vmax)
            if mag > 0:
                mag = ceil(mag)
            else:
                mag = floor(mag)

        if vmin is None:
            vmin = 10 ** (mag - 3)

        cbar_fig, cbar_ax = plt.subplots()
        if logscale:
            levels = log_levels(vmin, vmax)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=LogNorm()
            )
        else:
            levels = lin_levels(vmin, vmax, num=20)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=Normalize()
            )

        fig, ax = plt.subplots()
        for j, l in enumerate(levels):
            cntrs = find_contours(data, l)
            for c in cntrs:
                ax.fill(
                    x_intrp(c[:, 1]),
                    y_intrp(c[:, 0]),
                    color=cmap(j / len(levels)),
                    alpha=alpha,
                    zorder=1,
                )

        source = self.source_marker.to_crs(webmerc["init"])
        source.plot(ax=ax, marker="^", color="k", markersize=20, zorder=2)

        xlim = list(ax.get_xlim())
        ylim = list(ax.get_ylim())

        x_width = xlim[1] - xlim[0]
        y_height = ylim[1] - ylim[0]

        min_ax_width = int(1.5 * x_width)
        min_ax_height = int(1.5 * y_height)

        set_min_axes(ax, min_width=min_ax_width, min_height=min_ax_height)

        fig = set_figure_size(fig, ax)

        ax.get_xaxis().set_visible(False)
        ax.get_yaxis().set_visible(False)

        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.1)

        cbar = plt.colorbar(tmp, ax=ax, cax=cax)
        cbar.minorticks_off()
        # cbar.set_ticks([levels])
        cbar.set_label("Ash load (kg/m\u00B2)")
        plt.close(cbar_fig)

        ax = ax_ticks(ax, source.geometry[0].x, source.geometry[0].y)

        if basemap:
            ax = add_stamen_basemap(ax, stamen="background", zorder=0)
            (Narrow, ntext) = add_north_arrow(ax, zorder=11, fontsize=16)
            (scalebar, sbframe) = add_scale_bar(ax, segments=1)

        plt.draw()
        if save:
            plt.savefig(save_name, dpi=300, transparent=True, format="png")

        if show:
            plt.show()
        else:
            plt.close(fig)
        return (fig, ax, cbar)

    def contour_ashload_for_grain_class(self, grain_i, cntrs_levels="log", vmin=1e-2):

        self._check_valid_grain(grain_i)

        ashload = self.raster_ashload_for_grain_class(
            grain_i,
            vmin=1e-6,
            nodata=np.nan,
            masked=False,
            crs=webmerc["init"],
        )

        maxload = np.nanmax(ashload["ash_load"])
        mag = np.log10(maxload)
        if mag > 0:
            mag = ceil(mag)
        else:
            mag = floor(mag)
        vmax = nice_round_up(maxload, mag=10**mag)

        if vmin is None:
            vmin = 10 ** (mag - 3)

        if cntrs_levels == "log":
            levels = log_levels(vmin, vmax)
        else:
            levels = lin_levels(vmin, vmax)

        g = self._raster_contour(
            ashload["ash_load"], str(self.params.grains.diameter[grain_i]), levels
        )
        g.rename(columns={"contour": "load", "name": "grain_size"}, inplace=True)
        return g

    def plot_conc_for_grain_class(
        self, grain_i, logscale=True, vmin=1e-6, cmap=plt.cm.bone, basemap=False
    ):
        self._check_valid_grain(grain_i)
        Nz = self.params.output.Nz
        rows, cols = plot_rowscols(Nz + 1)
        fig, axes = plt.subplots(nrows=rows, ncols=cols, sharex=True, sharey=True)
        maxc = np.nanmax(self.Cz[:, :, :, grain_i])
        mag = np.log10(maxc)
        if mag > 0:
            mag = ceil(mag)
        else:
            mag = floor(mag)
        vmax = nice_round_up(maxc, mag=10**mag)

        cbar_fig, cbar_ax = plt.subplots()
        if logscale:
            levels = log_levels(vmin, vmax)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=LogNorm()
            )
        else:
            levels = lin_levels(vmin, vmax, num=20)
            tmp = cbar_ax.scatter(
                levels, np.ones_like(levels), c=levels, cmap=cmap, norm=Normalize()
            )

        x = self.x[:, grain_i] / 1e3
        y = self.y[:, grain_i] / 1e3
        for j, ax in enumerate(axes.reshape(-1)):
            if j < Nz:
                data = self.Cz[:, :, Nz - j - 1, grain_i]
                if np.nanmax(data) > vmin:
                    # # data = np.ma.masked_where(data <= vmin, data)
                    data[data <= vmin] = np.nan
                    if logscale:
                        CS = ax.contourf(
                            x,
                            y,
                            data,
                            levels,
                            locator=ticker.LogLocator(),
                            cmap=cmap,
                            origin="lower",
                        )
                    else:
                        CS = ax.contourf(
                            x,
                            y,
                            data,
                            levels,
                            cmap=cmap,
                            origin="lower",
                            vmin=vmin,
                            vmax=vmax,
                        )
                ax.set_title(
                    "z = {} m".format(self.params.output.altitudes[Nz - j - 1])
                )
            elif j == Nz:
                data = self.C0[:, :, grain_i]
                data[data <= 1e-6] = np.nan
                if np.nanmax(data) > vmin:
                    if logscale:
                        CS = ax.contourf(
                            x,
                            y,
                            data,
                            levels,
                            locator=ticker.LogLocator(),
                            cmap=cmap,
                            origin="lower",
                        )
                    else:
                        CS = ax.contourf(
                            x,
                            y,
                            data,
                            levels,
                            cmap=cmap,
                            origin="lower",
                            vmax=vmax,
                        )
                ax.set_title("z = 0 m")
            else:
                ax.axis("off")

        fig.subplots_adjust(right=0.75)
        cbar_ax = fig.add_axes([0.8, 0.2, 0.05, 0.7])

        cbar = fig.colorbar(tmp, cax=cbar_ax)
        cbar.minorticks_off()
        cbar.set_label("Ash concentration (kg/m\u00B3)")
        plt.close(cbar_fig)

        if basemap:
            ax = add_stamen_basemap(ax, stamen="background", zorder=0)
            (Narrow, ntext) = add_north_arrow(ax, zorder=11, fontsize=16)
            (scalebar, sbframe) = add_scale_bar(ax, segments=1)

        return (fig, ax, cbar)

    def plot_iso_conc_for_grain_class(self, grain_i, conc):
        self._check_valid_grain(grain_i)
        C = self.Cz[:, :, :, grain_i]

        if np.nanmax(C) < conc:
            message = (
                "In plotIsoConc, no concentration values in excess of"
                + " {}".format(conc)
            )
            print(message)
            return

        xyScale = self.params.model.xyScale[grain_i]

        z = self.params.output.altitudes

        verts, faces, _, _ = marching_cubes(
            np.where(C > conc, np.ones_like(C), np.zeros_like(C)),
            0.5,
            spacing=(1, 1, 1),
        )

        dx = self.x_dimless[1] - self.x_dimless[0]
        dy = self.y_dimless[1] - self.y_dimless[0]
        dz = z[1] - z[0]

        yv = (
            (verts[:, 0] * dx - np.pi)
            * xyScale
            * self.params.model.Lx[grain_i]
            / np.pi
            / 1e3
        )
        xv = (
            (verts[:, 1] * dy - np.pi)
            * xyScale
            * self.params.model.Ly[grain_i]
            / np.pi
            / 1e3
        )
        zv = verts[:, 2] * dz

        fig = plt.figure()
        ax = fig.add_subplot(111, projection="3d")

        ax.plot_trisurf(xv, yv, faces, zv, cmap="Spectral", lw=1)
        ax.set_xlabel("Easting (km)")
        ax.set_ylabel("Northing (km)")
        ax.set_zlabel("Altitude (m)")

    def plot_ashload(
        self,
        resolution=300.0,
        logscale=True,
        cmap=plt.cm.viridis,
        vmin=1e-3,
        nodata=-1,
        export_gtiff=False,
        export_name="AshLoad.tif",
    ):

        x, y, data = self.get_ashload(
            resolution=resolution,
            vmin=vmin / 10,
            nodata=nodata,
            export_gtiff=export_gtiff,
            export_name=export_name,
        )
        fig, ax = plt.subplots()

        maxc = np.nanmax(data)
        mag = np.log10(maxc)
        if mag > 0:
            mag = ceil(mag)
        else:
            mag = floor(mag)
        vmax = nice_round_up(maxc, mag=10**mag)
        x = x / 1e3
        y = y / 1e3
        data = np.ma.masked_where(data <= vmin, data)
        if logscale:
            levels = log_levels(vmin, vmax)
            CS = ax.contourf(
                x,
                y,
                data,
                levels,
                locator=ticker.LogLocator(),
                cmap=cmap,
                origin="lower",
            )
        else:
            levels = lin_levels(vmin, vmax, num=20)
            CS = ax.contourf(x, y, data, levels, cmap=cmap, origin="lower")

        ax.set_xlabel("Easting (km)")
        ax.set_ylabel("Northing (km)")
        cbar = fig.colorbar(CS)
        cbar.minorticks_off()
        cbar.set_ticks([levels])
        cbar.set_label("Ash load (kg/m\u00B2)")
        return (fig, ax, cbar)

    def folium_ashloads(self, savename):

        m = folium.Map(
            location=[self.params.source.latitude, self.params.source.longitude],
            zoom_start=13,
            control_scale=True,
            prefer_canvas=True,
            tiles="stamenterrain",
        )

        tiles = ["Stamen Terrain", "OpenStreetMap"]
        for tile in tiles:
            folium.TileLayer(tile).add_to(m)

        style_func = lambda x: {
            "weight": 0.1,
            "color": "black",
            "fillColor": cmap(x["properties"]["load"]),
            "fillOpacity": 0.5,
        }

        for grain_i in range(self.params.grains.bins):
            g = self.contour_ashload_for_grain_class(grain_i, vmin=1e-4)
            g = g.to_crs(webmerc["init"])
            g["geoid"] = g.index.astype(str)
            loads = g[["geoid", "load", "grain_size", "geometry"]]
            loads["grain_size"] = loads["grain_size"].astype(float) * 1e6
            loads["grain_size"] = loads["grain_size"].map("{0:.2f}".format)

            geo_str = loads.to_json()

            vmin = loads.load.min()
            vmax = loads.load.max()

            levels = log_steps(
                nice_round_down(vmin, mag=10 ** np.floor(np.log10(vmin))),
                nice_round_up(vmax),
                step=10,
            )

            cmap = cm.linear.viridis.to_step(
                data=loads["load"], index=levels, method="log", round_method="log10"
            )
            cmap.caption = "ash load (kg/m^2) for {0:.2f} micron grains".format(
                self.params.grains.diameter[grain_i] * 1e6
            )

            lm = folium.features.GeoJson(
                loads,
                style_function=style_func,
                control=True,
                name="Grain size = {0:.2f} microns".format(
                    self.params.grains.diameter[grain_i] * 1e6
                ),
                tooltip=folium.features.GeoJsonTooltip(
                    fields=["grain_size", "load"],
                    aliases=["Grain size (microns)", "Ash load (kg/m^2)"],
                    style=(
                        "background-color: white; color: #333333; font-family: arial; font-size: 12px; padding: 10px;"
                    ),
                    sticky=True,
                ),
            )

            m.add_child(lm)
            m.add_child(cmap)
            m.add_child(BindColormap(lm, cmap))

        folium.LayerControl().add_to(m)

        m.save(savename)



from functools import cache

import numpy as np
from numba import complex128, float64, int64, jit, njit

# from numba.pycc import CC
from numba.types import Tuple

from ..containers.cheb import ChebContainer_type
from ..containers.velocities import VelocityContainer_type
from ..met.met import MetData_type
from ..params.params import Parameters_type
from .getters import (
    Source_z_dimless_value,
    lower_dWsdz,
    lower_U,
    lower_V,
    lower_Ws,
    upper_dWsdz,
    upper_U,
    upper_V,
    upper_Ws,
)

# Cash-Karp tableau
C2 = 0.2
C3 = 0.3
C4 = 0.6
C5 = 1.0
C6 = 0.875
A21 = 0.2
A31 = 3.0 / 40.0
A32 = 9.0 / 40.0
A41 = 0.3
A42 = -0.9
A43 = 1.2
A51 = -11.0 / 54.0
A52 = 2.5
A53 = -70.0 / 27.0
A54 = 35.0 / 27.0
A61 = 1631.0 / 55296.0
A62 = 175.0 / 512.0
A63 = 575.0 / 13824.0
A64 = 44275.0 / 110592.0
A65 = 253.0 / 4096.0
B1 = 37.0 / 378.0
B2 = 0.0
B3 = 250.0 / 621.0
B4 = 125.0 / 594.0
B5 = 0.0
B6 = 512.0 / 1771.0
DB1 = B1 - 2825.0 / 27648.0
DB2 = 0.0
DB3 = B3 - 18575.0 / 48384.0
DB4 = B4 - 13525.0 / 55296.0
DB5 = -277.0 / 14336.0
DB6 = B6 - 0.25


@jit(
    complex128(
        float64,
        complex128,
        complex128,
        float64,
        float64,
        int64,
        Parameters_type,
        MetData_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
# @njit(cache=True)
def ode_homo(x, y0, y1, kx, ky, grain_i, parameters, Met):

    pi = np.pi
    Lx = parameters.model.Lx[grain_i]
    Ly = parameters.model.Ly[grain_i]
    Pe = parameters.model.Peclet_number
    Vratio = parameters.model.Velocity_ratio[grain_i]
    R = parameters.model.Diffusion_ratio

    U_scale = parameters.met.U_scale
    Ws_scale = parameters.met.Ws_scale[grain_i]

    z = -np.log(x)

    U = Met.wind_U_value(z, U_scale)
    V = Met.wind_V_value(z, U_scale)

    Ws = Met.settling_speed_for_grain_class_value(parameters, grain_i, z, Ws_scale)
    dz = 1e-4
    Ws1 = Met.settling_speed_for_grain_class_value(
        parameters, grain_i, z + dz, Ws_scale
    )
    dWsdz = (Ws1 - Ws) / dz

    t1 = (R * Ws / Pe / Vratio - 1.0) * y1 / x

    c2 = (
        1j * pi * (kx / Lx * U + ky / Ly * V)
        + pi * pi * Pe / Vratio * (kx * kx / Lx / Lx + ky * ky / Ly / Ly)
        - dWsdz
    )
    t2 = c2 * R / Pe / Vratio * y0 / x / x

    return t1 + t2


@jit(
    complex128[::1](
        float64,
        complex128[::1],
        float64,
        float64,
        complex128,
        int64,
        Parameters_type,
        MetData_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
# @njit(cache=True)
def ode(xt, yt, kx, ky, fxy_ij, grain_i, parameters, Met):

    dy = np.zeros_like(yt, dtype=np.complex128)

    z = -np.log(xt)
    fz = Source_z_dimless_value(z, parameters.source.Suzuki_k)

    dy[0] = yt[1]
    dy[1] = ode_homo(xt, yt[0], yt[1], kx, ky, grain_i, parameters, Met)
    dy[1] = dy[1] - fxy_ij * fz / xt / xt
    dy[2] = yt[3]
    dy[3] = ode_homo(xt, yt[2], yt[3], kx, ky, grain_i, parameters, Met)

    return dy


@jit(
    Tuple((complex128[::1], complex128[::1]))(
        float64,
        complex128[::1],
        complex128[::1],
        float64,
        int64,
        float64,
        float64,
        complex128,
        Parameters_type,
        MetData_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
# @njit(cache=True)
def rkCK(x_in, y_in, dy, h, grain_i, kx, ky, fxy_ij, parameters, Met):

    xt = x_in
    k1 = dy

    xt = x_in + C2 * h
    yt = y_in + A21 * k1 * h
    k2 = ode(xt, yt, kx, ky, fxy_ij, grain_i, parameters, Met)

    xt = x_in + C3 * h
    yt = y_in + (A31 * k1 + A32 * k2) * h
    k3 = ode(xt, yt, kx, ky, fxy_ij, grain_i, parameters, Met)

    xt = x_in + C4 * h
    yt = y_in + (A41 * k1 + A42 * k2 + A43 * k3) * h
    k4 = ode(xt, yt, kx, ky, fxy_ij, grain_i, parameters, Met)

    xt = x_in + C5 * h
    yt = y_in + (A51 * k1 + A52 * k2 + A53 * k3 + A54 * k4) * h
    k5 = ode(xt, yt, kx, ky, fxy_ij, grain_i, parameters, Met)

    xt = x_in + C6 * h
    yt = y_in + (A61 * k1 + A62 * k2 + A63 * k3 + A64 * k4 + A65 * k5) * h
    k6 = ode(xt, yt, kx, ky, fxy_ij, grain_i, parameters, Met)

    # Accumulate increments with proper weights.
    y_out = y_in + h * (B1 * dy + B3 * k3 + B4 * k4 + B6 * k6)
    y_err = h * (DB1 * dy + DB3 * k3 + DB4 * k4 + DB5 * k5 + DB6 * k6)

    return y_out, y_err


@jit(
    Tuple((float64, complex128[::1], float64))(
        float64,
        complex128[::1],
        complex128[::1],
        float64,
        int64,
        float64,
        float64,
        complex128,
        Parameters_type,
        MetData_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
# @njit(cache=True)
def rkstep(x, y, dy, h_in, grain_i, kx, ky, fxy_ij, parameters, Met):

    rtol = parameters.solver.rtol
    maxStep = parameters.solver.maxStep
    meps = parameters.solver.meps

    h = h_in
    while True:
        yt, y_err = rkCK(x, y, dy, h, grain_i, kx, ky, fxy_ij, parameters, Met)

        y_scal = np.absolute(yt) + np.absolute(h * dy) + meps
        errmax = np.max(np.absolute(y_err) / np.absolute(y_scal)) / rtol

        if errmax <= 1:
            break

        htemp = 0.9 * h * (np.power(errmax, -0.25))
        htemp = np.maximum(np.abs(htemp), 0.1 * np.abs(h))
        h = np.abs(htemp) * np.sign(h)

    h_out = h

    if errmax > 1.89e-4:
        h_next = 0.9 * h / (np.power(errmax, 0.2))
    else:
        h_next = 5.0 * h  # Numerical Recipes criteria

    x_out = x + h_out
    y_out = yt

    h_next = np.minimum(h_next, maxStep)

    return x_out, y_out, h_next


@jit(
    Tuple((complex128[::1], complex128[::1], float64))(
        float64,
        float64,
        complex128[::1],
        complex128[::1],
        float64,
        int64,
        float64,
        float64,
        complex128,
        Parameters_type,
        MetData_type,
    ),
    nopython=True,
    cache=True,
    fastmath=True,
)
# @njit(cache=True)
def integrateTo(x0, x1, y, dy, h_in, grain_i, kx, ky, fxy_ij, parameters, Met):

    x = x0
    h = h_in
    h_next = h_in
    while x < x1:
        x, y, h_next = rkstep(x, y, dy, h, grain_i, kx, ky, fxy_ij, parameters, Met)

        dy = ode(x, y, kx, ky, fxy_ij, grain_i, parameters, Met)

        if x + h_next > x1:
            h = x1 - x

    return y, dy, h_next



from collections import OrderedDict

import numpy as np
from numba import boolean, float64, int64
from numba.experimental import jitclass

solver_spec = OrderedDict()
solver_spec["domX"] = float64  # dimensionless domain length in x
solver_spec["domY"] = float64  # dimensionless domain length in y
solver_spec["minN_log2"] = int64  # log2 minimum number of Chebyshev points
solver_spec["minN"] = int64  # minimum number of Chebyshev points
solver_spec["maxN_log2"] = int64  # log2 maximum number of Chebyshev points
solver_spec["maxN"] = int64  # maximum number of Chebyshev points
solver_spec["Nx_log2"] = int64  # log2 of number of points in x
solver_spec["Nx"] = int64  # number of points in x
solver_spec["Ny_log2"] = int64  # log2 of number of points in y
solver_spec["Ny"] = int64  # number of points in y
solver_spec["chebIts"] = int64  # number of chebyshev iterates
solver_spec["epsilon"] = float64  # tolerance for converged spectral series
solver_spec["meps"] = float64  # Machine epsilon
solver_spec["rtol"] = float64  # tolerance for Runge-Kutta integration
solver_spec["maxStep"] = float64  # max step for Runge-Kutta integration
solver_spec["rk"] = boolean # use rk solver


@jitclass(solver_spec)
class SolverParameters:
    def __init__(
        self,
        domX=1.5,
        domY=1.5,
        minN_log2=4,
        maxN_log2=8,
        Nx_log2=8,
        Ny_log2=8,
        epsilon=1e-8,
        rtol=1e-8,
        maxStep=0.01,
        rk=False
    ):
        self.meps = np.finfo(np.float64).eps

        if domX < 0:
            raise ValueError("In SolverParameters, must have domX>0")
        self.domX = np.float64(domX)  # Dimensionless domain size in x

        if domY < 0:
            raise ValueError("In SolverParameters, must have domY>0")
        self.domY = np.float64(domY)  # Dimensionless domain size in y

        if minN_log2 < 0:
            raise ValueError("In SolverParameters, must have minN_log2>0")
        self.minN_log2 = np.int64(minN_log2)  # Minimum z-resolution (log2)
        self.minN = 2**self.minN_log2

        if maxN_log2 < 0:
            raise ValueError("In SolverParameters, must have maxN_log2>0")
        self.maxN_log2 = np.int64(maxN_log2)  # Maximum z-resolution (log2)
        self.maxN = 2**self.maxN_log2

        if minN_log2 > maxN_log2:
            raise ValueError("In SolverParameters, must have minN_log2 < maxN_log2")

        if Nx_log2 < 0:
            raise ValueError("In SolverParameters, must have Nx_log2>0")
        self.Nx_log2 = np.int64(Nx_log2)  # x-resolution (log2)
        self.Nx = 2**self.Nx_log2

        if Ny_log2 < 0:
            raise ValueError("In SolverParameters, must have Ny_log2>0")
        self.Ny_log2 = np.int64(Ny_log2)  # y-resolution (log2)
        self.Ny = 2**self.Ny_log2

        self.chebIts = self.maxN_log2 - self.minN_log2 + 1

        if epsilon < 0:
            raise ValueError("In SolverParameters, must have epsilon>0")
        if epsilon < 10 * self.meps:
            raise ValueError(
                "In SolverParameters, " + "must have epsilon >= 10*machine epsilon"
            )
        self.epsilon = np.float64(epsilon)

        if rtol < 0:
            raise ValueError("In SolverParameters, must have rtol>0")
        self.rtol = np.float64(rtol)  # tolerance for Runge-Kutta integration

        if maxStep <= 0:
            raise ValueError("In SolverParameters, must have maxStep>0")
        if maxStep >= 1:
            raise ValueError("In SolverParameters, must have maxStep<1")
        self.maxStep = np.float64(maxStep)  # max step for Runge-Kutta integration

        if rk:
            self.rk = True
        else:
            self.rk = False

    def describe(self):
        print("Solver parameters for AshDisperse")
        print("  Dimensionless domain size in x, domX = ", self.domX)
        print("  Dimensionless domain size in y, domY = ", self.domY)
        print(
            "  Minimum resolution in z, minN = ",
            self.minN,
            " (minN_log2 = ",
            self.minN_log2,
            ")",
        )
        print(
            "  Maximum resolution in z, maxN = ",
            self.maxN,
            " (maxN_log2 = ",
            self.maxN_log2,
            ")",
        )
        print("  Number of Chebyshev iterates = ", self.chebIts)
        print("  Tolerance for Chebyshev series, epsilon = ", self.epsilon)
        print("  Resolution in x, Nx = ", self.Nx, " (Nx_log2 = ", self.Nx_log2, ")")
        print("  Resolution in y, Ny = ", self.Ny, " (Ny_log2 = ", self.Ny_log2, ")")
        print("********************")


# pylint: disable=E1101
SolverParameters_type = SolverParameters.class_type.instance_type



import numpy as np
from numba import complex128, float64, njit, prange
from numba.types import Tuple

import ashdisperse.core.core as core
import ashdisperse.spectral.cheb as cheb
from ashdisperse.containers import ChebContainer_type, VelocityContainer_type
from ashdisperse.met.met import MetData_type
from ashdisperse.params import Parameters_type


@njit(
    Tuple((complex128[::1], complex128[:, ::1]))(
        float64, float64, complex128[::1], Parameters_type, MetData_type
    ),
    cache=True,
    parallel=False,
    fastmath=True,
)
def ade_mode_rk_solve(kx, ky, fxy_ij, parameters, Met):

    z = np.zeros((parameters.output.Nz), dtype=np.float64)
    z = parameters.output.altitudes
    z = z / parameters.source.PlumeHeight

    conc_0_ft = np.zeros((parameters.grains.bins), dtype=np.complex128)
    conc_z_ft = np.zeros((z.size, parameters.grains.bins), dtype=np.complex128)

    for grain_i in range(parameters.grains.bins):
        C = core.rkBVP(kx, ky, fxy_ij[grain_i], grain_i, parameters, Met)

        conc_0_ft[grain_i] = C[-1]
        conc_z_ft[:, grain_i] = C[::-1]

    return conc_0_ft, conc_z_ft


@njit(
    Tuple((complex128[::1], complex128[:, ::1]))(
        float64,
        float64,
        complex128[::1],
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
    ),
    cache=True,
    parallel=False,
    fastmath=True,
)
def ade_mode_solve(kx, ky, fxy_ij, ChebContainer, parameters, VelocityContainer):

    z = np.zeros((parameters.output.Nz), dtype=np.float64)
    z = parameters.output.altitudes
    z = z / parameters.source.PlumeHeight
    Nlow = z[z <= 1].size

    conc_0_ft = np.zeros((parameters.grains.bins), dtype=np.complex128)
    conc_z_ft = np.zeros((z.size, parameters.grains.bins), dtype=np.complex128)

    for grain_i in range(parameters.grains.bins):
        """Lower domain.
        Solve for:
          inhomogeneous part (called p1) with null boundary conditions
          homogeneous part (called r1) with bcs r1(1) = 1, r1(-1) = 0
          homogeneous part (called l1) with bcs l1(1) = 0, l1(-1) = 1
          with _c for coefficients, _v for values)
        """
        p1_c, l1_c, r1_c = core.LowerODE(
            kx,
            ky,
            fxy_ij[grain_i],
            np.int64(grain_i),
            ChebContainer,
            parameters,
            VelocityContainer,
        )

        dp1_p1 = cheb.cheb_dif_p1(p1_c)
        dr1_p1 = cheb.cheb_dif_p1(r1_c)
        dl1_p1 = cheb.cheb_dif_p1(l1_c)

        dp1_m1 = cheb.cheb_dif_m1(p1_c)
        dr1_m1 = cheb.cheb_dif_m1(r1_c)
        dl1_m1 = cheb.cheb_dif_m1(l1_c)

        """Upper domain.
             Solve for:
               inhomogeneous part (called p2) with null boundary conditions
               homogeneous part (called r2) with bcs r2(1) = 1, r2(-1) = 0
               with _c for coefficients, _v for values)
        """
        l2_c = core.UpperODE(
            kx, ky, np.int64(grain_i), ChebContainer, parameters, VelocityContainer
        )

        dl2_m1 = cheb.cheb_dif_m1(l2_c)

        a = dl1_m1
        b = dr1_m1
        c = dl1_p1
        d = dr1_p1 - dl2_m1

        b0 = -dp1_m1
        b1 = -dp1_p1

        delta = a * d - b * c

        match0 = (b0 * d - b1 * b) / delta
        match1 = (b1 * a - b0 * c) / delta

        Cheb_r1 = np.zeros(
            (parameters.output.Cheb_lower.shape[0], r1_c.size), dtype=np.complex128
        )
        Cheb_l1 = np.zeros(
            (parameters.output.Cheb_lower.shape[0], l1_c.size), dtype=np.complex128
        )
        Cheb_p1 = np.zeros(
            (parameters.output.Cheb_lower.shape[0], p1_c.size), dtype=np.complex128
        )

        Cheb_r1[:, : r1_c.size] = parameters.output.Cheb_lower[:, : r1_c.size]
        Cheb_l1[:, : l1_c.size] = parameters.output.Cheb_lower[:, : l1_c.size]
        Cheb_p1[:, : p1_c.size] = parameters.output.Cheb_lower[:, : p1_c.size]

        r1_v = Cheb_r1 @ r1_c
        l1_v = Cheb_l1 @ l1_c
        p1_v = Cheb_p1 @ p1_c

        conc_z_ft[:Nlow, grain_i] = p1_v + match0 * l1_v + match1 * r1_v

        Cheb_l2 = np.zeros(
            (parameters.output.Cheb_upper.shape[0], l2_c.size), dtype=np.complex128
        )
        Cheb_l2[:, : l2_c.size] = parameters.output.Cheb_upper[:, : l2_c.size]
        l2_v = Cheb_l2 @ l2_c

        conc_z_ft[Nlow:, grain_i] = match1 * l2_v

        conc_0_ft[grain_i] = match0  # (b0*d - b1*b)/delta

    return conc_0_ft, conc_z_ft


@njit(
    Tuple((complex128[:, :, ::1], complex128[:, :, :, ::1]))(
        float64[::1],
        float64[::1],
        complex128[:, :, ::1],
        ChebContainer_type,
        Parameters_type,
        VelocityContainer_type,
    ),
    parallel=True,
    cache=True,
    fastmath=True,
)
def ade_ft_system(kx, ky, fxy_f, cheby, params, velocities):

    Nx = kx.size
    Ny = ky.size

    conc_0_fft = np.zeros((Ny, Nx, params.grains.bins), dtype=np.complex128)
    conc_z_fft = np.zeros(
        (Ny, Nx, params.output.Nz, params.grains.bins), dtype=np.complex128
    )

    # do kx = ky = 0:
    conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve(
        0.0, 0.0, fxy_f[0, 0, :], cheby, params, velocities
    )
    conc_0_fft[0, 0, :] = conc_0_mode_fft
    conc_z_fft[0, 0, :, :] = conc_z_mode_fft

    # do kx = 0, ky = 1 ... Ny/2-1, -Ny/2
    # and we get kx = 0, ky = -Ny/2+1, ... , -1 for free by conjugation
    for j in prange(1, Ny // 2 + 1):

        conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve(
            0.0, ky[j], fxy_f[j, 0, :], cheby, params, velocities
        )

        conc_0_fft[j, 0, :] = conc_0_mode_fft
        conc_0_fft[Ny - j, 0, :] = np.conj(conc_0_mode_fft)

        conc_z_fft[j, 0, :, :] = conc_z_mode_fft
        conc_z_fft[Ny - j, 0, :, :] = np.conj(conc_z_mode_fft)

    # do ky = 0, kx = 1 ... Nx/2-1, -Nx/2
    # and we get ky = 0, kx = -Nx/2+1, ... , -1 for free by conjugation
    for i in prange(1, Nx // 2 + 1):

        conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve(
            kx[i], 0.0, fxy_f[0, i, :], cheby, params, velocities
        )

        conc_0_fft[0, i, :] = conc_0_mode_fft
        conc_0_fft[0, Nx - i, :] = np.conj(conc_0_mode_fft)

        conc_z_fft[0, i, :, :] = conc_z_mode_fft
        conc_z_fft[0, Nx - i, :, :] = np.conj(conc_z_mode_fft)

    # Do first quadrant, kx = 1 ... Nx/2-1, ky = 1 .. Ny/2-1
    # and get the 4th quadrant for free by conjugation.
    # Also do second quadrant, kx = -Nx/2+1 ... -1, ky = 1 .. Ny/2-1
    # and get the third quadrant for free by conjugation.
    for i in prange(1, Nx // 2):
        for j in range(1, Ny // 2):

            conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve(
                kx[i], ky[j], fxy_f[j, i, :], cheby, params, velocities
            )
            conc_0_fft[j, i, :] = conc_0_mode_fft
            conc_z_fft[j, i, :, :] = conc_z_mode_fft

            conc_0_mode_fft, conc_z_mode_fft = ade_mode_solve(
                kx[Nx - i], ky[j], fxy_f[j, Nx - i, :], cheby, params, velocities
            )
            conc_0_fft[j, Nx - i, :] = conc_0_mode_fft
            conc_z_fft[j, Nx - i, :, :] = conc_z_mode_fft

            conc_0_fft[Ny - j, Nx - i, :] = np.conj(conc_0_fft[j, i, :])
            conc_0_fft[Ny - j, i, :] = np.conj(conc_0_fft[j, Nx - i, :])

            conc_z_fft[Ny - j, Nx - i, :, :] = np.conj(conc_z_fft[j, i, :, :])
            conc_z_fft[Ny - j, i, :, :] = np.conj(conc_z_fft[j, Nx - i, :, :])

    return conc_0_fft, conc_z_fft


@njit(
    Tuple((complex128[:, :, ::1], complex128[:, :, :, ::1]))(
        float64[::1],
        float64[::1],
        complex128[:, :, ::1],
        Parameters_type,
        MetData_type,
    ),
    parallel=True,
    cache=True,
    fastmath=True,
)
def ade_ft_system_rk(kx, ky, fxy_f, params, Met):

    Nx = kx.size
    Ny = ky.size

    conc_0_fft = np.zeros((Ny, Nx, params.grains.bins), dtype=np.complex128)
    conc_z_fft = np.zeros(
        (Ny, Nx, params.output.Nz, params.grains.bins), dtype=np.complex128
    )

    # do kx = ky = 0:
    conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
        0.0, 0.0, fxy_f[0, 0, :], params, Met
    )
    conc_0_fft[0, 0, :] = conc_0_mode_fft
    conc_z_fft[0, 0, :, :] = conc_z_mode_fft

    # do kx = 0, ky = 1 ... Ny/2-1, -Ny/2
    # and we get kx = 0, ky = -Ny/2+1, ... , -1 for free by conjugation
    for j in prange(1, Ny // 2 + 1):

        conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
            0.0, ky[j], fxy_f[j, 0, :], params, Met
        )

        conc_0_fft[j, 0, :] = conc_0_mode_fft
        conc_0_fft[Ny - j, 0, :] = np.conj(conc_0_mode_fft)

        conc_z_fft[j, 0, :, :] = conc_z_mode_fft
        conc_z_fft[Ny - j, 0, :, :] = np.conj(conc_z_mode_fft)

    # do ky = 0, kx = 1 ... Nx/2-1, -Nx/2
    # and we get ky = 0, kx = -Nx/2+1, ... , -1 for free by conjugation
    for i in prange(1, Nx // 2 + 1):

        conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
            kx[i], 0.0, fxy_f[0, i, :], params, Met
        )

        conc_0_fft[0, i, :] = conc_0_mode_fft
        conc_0_fft[0, Nx - i, :] = np.conj(conc_0_mode_fft)

        conc_z_fft[0, i, :, :] = conc_z_mode_fft
        conc_z_fft[0, Nx - i, :, :] = np.conj(conc_z_mode_fft)

    # Do first quadrant, kx = 1 ... Nx/2-1, ky = 1 .. Ny/2-1
    # and get the 4th quadrant for free by conjugation.
    # Also do second quadrant, kx = -Nx/2+1 ... -1, ky = 1 .. Ny/2-1
    # and get the third quadrant for free by conjugation.
    for i in prange(1, Nx // 2):
        for j in range(1, Ny // 2):

            conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
                kx[i], ky[j], fxy_f[j, i, :], params, Met
            )
            conc_0_fft[j, i, :] = conc_0_mode_fft
            conc_z_fft[j, i, :, :] = conc_z_mode_fft

            conc_0_mode_fft, conc_z_mode_fft = ade_mode_rk_solve(
                kx[Nx - i], ky[j], fxy_f[j, Nx - i, :], params, Met
            )
            conc_0_fft[j, Nx - i, :] = conc_0_mode_fft
            conc_z_fft[j, Nx - i, :, :] = conc_z_mode_fft

            conc_0_fft[Ny - j, Nx - i, :] = np.conj(conc_0_fft[j, i, :])
            conc_0_fft[Ny - j, i, :] = np.conj(conc_0_fft[j, Nx - i, :])

            conc_z_fft[Ny - j, Nx - i, :, :] = np.conj(conc_z_fft[j, i, :, :])
            conc_z_fft[Ny - j, i, :, :] = np.conj(conc_z_fft[j, Nx - i, :, :])

    return conc_0_fft, conc_z_fft



from ashdisperse.ashdisperse import setup, solve
from ashdisperse.core import AshDisperseResult
from ashdisperse.interface import load_inputs, save_inputs, set_met, set_parameters
from ashdisperse.met import load_met, save_met, wind_plot
from ashdisperse.params import (
    copy_parameters,
    load_parameters,
    save_parameters,
    update_parameters,
)
from ashdisperse.version import __version__

__author__ = "Mark Woodhouse"



import argparse
from ashdisperse.config import set_default_threads
from ashdisperse.config.config import set_threading_layer
from ashdisperse.interface import set_parameters, set_met
from ashdisperse.params import load_parameters
from ashdisperse.ashdisperse import solve
import pathlib
from datetime import datetime

print("Running AshDisperse")
now = datetime.utcnow()
run_start = now.strftime("%Y%m%d%H%M")

parser = argparse.ArgumentParser()
parser.add_argument("--parameters", "-p", type=str, help="Parameters file name.")
parser.add_argument("--met_source", "-m", type=str,
                    choices=['interface', 'gfs', 'ecmwf', 'local'], help="Source of met data.")
parser.add_argument("-d", "--met_datetime", type=str, help="Date-time string for met data, in format yyyy-mm-dd hh:mm")

args = parser.parse_args()

if args.met_source in {'gfs', 'ecmwf'} and args.met_datetime is None:
    parser.error("met_datetime is required when met_source is {}".format(args.met_source))

if args.parameters is not None:
    params_file = pathlib.Path(args.parameters)
    if not params_file.exists():
        raise FileExistsError('Parameters file {} does not exist'.format(params_file))
    if not params_file.is_file():
        raise FileNotFoundError('Input parameters {} is not a file'.format(args.parameters))

    params = load_parameters(params_file)
else:
    params = set_parameters()

if args.met_source is not None:
    met_source = args.met_source
    met_datetime = args.met_datetime
    met_datetime_str = met_datetime.replace('-','').replace(':','').replace(' ','')

    params, met = set_met(params, source=met_source, datetime=met_datetime)

    save_name = 'AshDisperse_{name}_{met_source}_{met_date}_{run_start}'.format(name=params.source.name, 
                                                                            met_source=met_source,
                                                                            met_date=met_datetime_str,
                                                                            run_start=run_start)
else:
    params, met = set_met(params)
    save_name = 'AshDisperse_{name}_{run_start}'.format(name=params.source.name, run_start=run_start)

set_threading_layer(thread_layer='omp')
set_default_threads()
result = solve(params, met, timer=True)

for grain_i in range(0, params.grains.bins):
    result.plot_ashload_for_grain_class(0, vmin=1e-2, save_name=save_name)

# for grain_i in range(0, params.grains.bins):
#     result.plotSettlingFlux(grain_i)
#     result.plotConc(grain_i)
#     result.plotIsoConc(grain_i, 1e-4)

# _ = result.getAshLoad(resolution=500., vmin=1e-4, nodata=-1,
#                       export_gtiff=True, export_name='AshLoad.tif')



# -*- coding: utf-8 -*-
"""Grain parameters for AshDisperse.

Defines the GrainParameters class, containing attributes specifying grains
classes and methods for amending these, and a printer for this object.

Example:
    grains = GrainParameters()
    grains.add_grains(1e-4,1400,0.5)
    grains.add_grains(1e-3,1400,0.5)

    print_grain_parameters(grains)
"""

from collections import OrderedDict

import numpy as np
from numba import float64, int64
from numba.experimental import jitclass
from numba.typed import List
from numba.types import ListType

grain_spec = OrderedDict()
grain_spec["bins"] = int64  # Number of grain classes
grain_spec["diameter"] = ListType(float64)  # Grain sizes
grain_spec["density"] = ListType(float64)  # Grain densities
grain_spec["proportion"] = ListType(float64)  # Grain proportions
"""OrderedDict: GrainParameters attribute type specification.

This specification is required for numba jitting.
"""


@jitclass(grain_spec)
class GrainParameters:
    """Grain parameters required for AshDisperse.

    Defines the GrainParameters class, containing attributes specifying grains
    classes and methods for amending these.

    Attributes:
        bins: An integer count of the number of grain classes.
        diameter: A list of floats containing the diameter of each grain class.
        density: A list of floats containing the density of each grain class.
        proportion: A list of floats containing the proportion of each grain
                    class.
    """

    def __init__(self):
        """Initialize of GrainParameters to empty lists.

        Use add_grain to add grains.
        """
        self.bins = 0
        self.diameter = List.empty_list(np.float64)
        self.density = List.empty_list(np.float64)
        self.proportion = List.empty_list(np.float64)

    def fromvalue(self, diameter, density, proportion):
        self.add_grain(diameter, density, proportion)
        return

    def fromlists(self, diameters, densities, proportions):
        "Initialize grain parameters from lists"
        if diameters.size != densities.size:
            raise ValueError("Size of diameters and densities not equal")
        if diameters.size != proportions.size:
            raise ValueError("Size of diameters and proportions not equal")
        if proportions.size != densities.size:
            raise ValueError("Size of densities and proportions not equal")

        for j, (d, rho, p) in enumerate(zip(diameters, densities, proportions)):
            self.add_grain(d, rho, p)

        return

    def add_grain(self, diameter, density, proportion):
        """Add a grain class to the GrainParameters object.

        A grain class requires a diameters, density and proportion.
        The proportions must sum to one.
        Adding a grain increments the number of classes through the bins
        attribute.

        Args:
            diameter: [float64] The diameter of the grain class;
                must be positive.
            density: [float64] The density of the grain class;
                must be positive.
            proportion: [float64] The proportion of the total mass in this
                grain class; must be in the range (0,1].
        """
        if diameter < 0:
            raise ValueError("Grain diameter must be positive")
        if density < 0:
            raise ValueError("Grain density must be positive")
        if proportion < 0:
            raise ValueError("Grain proportion must be in [0,1]")
        if proportion > 1:
            raise ValueError("Grain proportion must be in [0,1]")
        self.diameter.append(np.float64(diameter))
        self.density.append(np.float64(density))
        self.proportion.append(np.float64(proportion))
        self.bins += 1

    def remove_grain(self, grain_i):
        """Remove a grain class to the GrainParameters object.

        The grain class with a given index is removed from the GrainParameters
        object.
        Removing a grain decrements the number of classes through the bins
        attribute.

        Args:
            grain_i: [int] The index of the grain class.  Must be less than the
                           number of grain classes.
        """
        if grain_i < 0:
            raise ValueError("index to grain class must be positive")
        if grain_i >= self.bins:
            raise ValueError("index must in range [0,{}]".format(self.bins - 1))
        self.diameter.pop(grain_i)
        self.density.pop(grain_i)
        self.proportion.pop(grain_i)
        self.bins -= 1

    def validate(self):
        """Validate grain classes in GrainParameters object."""
        retval = 1
        if len(self.diameter) != self.bins:
            print(
                "In GrainParameters: number of grain diameter values (",
                len(self.diameter),
                ") must equal number of grain classes (",
                self.bins,
                ")",
            )
            retval = 0
        if len(self.density) != self.bins:
            print(
                "In GrainParameters: number of grain density values must "
                + "equal number of grain classes"
            )
            retval = 0
        if len(self.proportion) != self.bins:
            print(
                "In GrainParameters: number of grain proportion values "
                + "must equal number of grain classes"
            )
            retval = 0
        sum_prop = np.sum(np.array(list(self.proportion), dtype=float64), dtype=float64)
        if np.abs(sum_prop - 1.0) > np.finfo(np.float64).eps:
            print(
                "In GrainParameters: the sum of the proportions for the "
                + "grain classes must be unity"
            )
            retval = 0
        return retval

    def describe(self):
        """Describe the grain classes in GrainParameters object."""
        print("Grain parameters for AshDisperse")
        print("  Number of grain classes, N_grains = ", self.bins)
        for j in range(self.bins):
            print("  Grain class ", j + 1)
            print("    Grain diameter = ", self.diameter[j], " m")
            print("    Grain density = ", self.density[j], " kg/m^3")
            print("    proportion = ", self.proportion[j])
        print("********************")


# pylint: disable=E1101
GrainParameters_type = GrainParameters.class_type.instance_type



import datetime
import os
from collections import OrderedDict

import numpy as np
from numba import optional
from numba.experimental import jitclass

from ..queryreport import is_valid_lat, is_valid_lon
from ..utilities import latlon_point_to_utm_code
from .grain_params import GrainParameters, GrainParameters_type
from .met_params import MetParameters, MetParameters_type
from .model_params import ModelParameters, ModelParameters_type
from .output_params import OutputParameters, OutputParameters_type
from .physical_params import PhysicalParameters, PhysicalParameters_type
from .solver_params import SolverParameters, SolverParameters_type
from .source_params import SourceParameters, SourceParameters_type

param_spec = OrderedDict()
param_spec["solver"] = optional(SolverParameters_type)
param_spec["grains"] = optional(GrainParameters_type)
param_spec["source"] = optional(SourceParameters_type)
param_spec["physical"] = optional(PhysicalParameters_type)
param_spec["met"] = optional(MetParameters_type)
param_spec["model"] = optional(ModelParameters_type)
param_spec["output"] = optional(OutputParameters_type)


@jitclass(param_spec)
class Parameters:
    def __init__(self):
        self.solver = None
        self.grains = None
        self.source = None
        self.physical = None
        self.met = None
        self.model = None
        self.output = None

    def describe(self):
        print("AshDisperse parameters")
        self.source.describe()
        self.grains.describe()
        self.solver.describe()
        self.physical.describe()
        # self.met.describe()
        self.model.describe()
        self.output.describe()


# pylint: disable=E1101
Parameters_type = Parameters.class_type.instance_type


def copy_parameters(A):
    new_A = Parameters()
    new_A.solver = SolverParameters(
        domX=A.solver.domX,
        domY=A.solver.domY,
        minN_log2=A.solver.minN_log2,
        maxN_log2=A.solver.maxN_log2,
        epsilon=A.solver.epsilon,
        Nx_log2=A.solver.Nx_log2,
        Ny_log2=A.solver.Ny_log2,
    )
    new_A.grains = GrainParameters()
    for j in range(A.grains.bins):
        new_A.grains.add_grain(
            A.grains.diameter[j], A.grains.density[j], A.grains.proportion[j]
        )
    new_A.source = SourceParameters(
        A.source.latitude,
        A.source.longitude,
        A.source.utmcode,
        A.source.radius,
        A.source.PlumeHeight,
        A.source.MER,
        A.source.duration,
        A.source.Suzuki_k,
        None,
    )
    new_A.physical = PhysicalParameters(
        A.physical.Kappa_h, A.physical.Kappa_v, A.physical.g, A.physical.mu
    )
    new_A.met = MetParameters(A.met.U_scale, A.met.Ws_scale)
    new_A.model = ModelParameters()
    new_A.model.from_values(
        A.model.SettlingScale,
        A.model.Velocity_ratio,
        A.model.xyScale,
        A.model.Lx,
        A.model.Ly,
        A.model.cScale,
        A.model.QScale,
        A.model.Peclet_number,
        A.model.Diffusion_ratio,
        A.model.sigma_hat,
        A.model.sigma_hat_scale,
    )
    new_A.output = OutputParameters(A.output.start, A.output.stop, A.output.step)
    new_A.output.set_altitudes()
    new_A.output.ChebMats(A.solver.maxN, A.source.PlumeHeight)
    return new_A


def update_parameters(
    A,
    met,
    domX=None,
    domY=None,
    minN_log2=None,
    maxN_log2=None,
    epsilon=None,
    Nx_log2=None,
    Ny_log2=None,
    grains=None,
    latitude=None,
    longitude=None,
    radius=None,
    PlumeHeight=None,
    MER=None,
    duration=None,
    Suzuki_k=None,
    Suzuki_peak=None,
    Kappa_h=None,
    Kappa_v=None,
    g=None,
    mu=None,
    U_scale=None,
    Ws_scale=None,
    start=None,
    stop=None,
    step=None,
):

    # solver
    if domX is not None:
        A.solver.domX = domX
    if domY is not None:
        A.solver.domY = domY
    if minN_log2 is not None:
        A.solver.minN_log2 = minN_log2
        A.solver.minN = 2**minN_log2
    if maxN_log2 is not None:
        A.solver.maxN_log2 = maxN_log2
        A.solver.maxN = 2**maxN_log2
    A.solver.chebIts = A.solver.maxN_log2 - A.solver.minN_log2 + 1
    if epsilon is not None:
        A.solver.epsilon = epsilon
    if Nx_log2 is not None:
        A.solver.Nx_log2 = Nx_log2
        A.solver.Nx = 2**Nx_log2
    if Ny_log2 is not None:
        A.solver.Ny_log2 = Ny_log2
        A.solver.Ny = 2**Ny_log2

    # grains
    if grains is not None:
        if not isinstance(grains, list):
            raise ValueError(
                "in update_parameters, grains must be a list of dicts; "
                + "received {}".format(grains)
            )
        if not all(
            "class" in g and "diameter" in g and "density" in g and "proportion" in g
            for g in grains
        ):
            raise ValueError(
                "in update parameters, grains must be a list of dicts with "
                + "dict containing the keys 'class', 'diameter', 'density',"
                + " 'proportion'; received {}".format(grains)
            )
        grains = sorted(grains, key=lambda p: p["proportion"], reverse=True)
        for grain in grains:
            grain_i = grain["class"] - 1
            Diameter = grain["diameter"]
            Density = grain["density"]
            proportion = grain["proportion"]
            if grain_i < A.grains.bins:
                if proportion > 0:
                    A.grains.diameter[grain_i] = Diameter
                    A.grains.density[grain_i] = Density
                    A.grains.proportion[grain_i] = proportion
                else:
                    A.grains.remove_grain(grain_i)
            else:
                A.grains.add_grain(Diameter, Density, proportion)
        A.grains.validate()

    # source
    if latitude is not None:
        if is_valid_lat(latitude):
            A.source.latitude = np.float64(latitude)
        else:
            raise ValueError(
                "In update_parameters, latitude must be a valid latitude in "
                + " decimal degrees"
            )
    if longitude is not None:
        if is_valid_lon(longitude):
            A.source.longitude = np.float64(longitude)
        else:
            raise ValueError(
                "In update_parameters, latitude must be a valid longitude in "
                + " decimal degrees"
            )
    if (latitude is not None) or (longitude is not None):
        A.source.utmcode = latlon_point_to_utm_code(
            A.source.latitude, A.source.longitude
        )

    if radius is not None:
        if radius < 0:
            raise ValueError("In update_parameters, radius must be positive")
        A.source.radius = np.float64(radius)

    if PlumeHeight is not None:
        if PlumeHeight < 0:
            raise ValueError("In update_parameters, PlumeHeight must be positive")
        A.source.PlumeHeight = np.float64(PlumeHeight)

    if MER is not None:
        if MER < 0:
            raise ValueError("In update_parameters, MER must be positive")
        A.source.MER = np.float64(MER)

    if duration is not None:
        if duration < 0:
            raise ValueError("In update_parameters, duration must be positive")
        A.source.duration = np.float64(duration)

    if Suzuki_k is not None:
        if Suzuki_k < 1:
            raise ValueError("In update_parameters, must have Suzuki_k>=1")
        A.source.Suzuki_k = np.float64(Suzuki_k)
        A.source.peak_from_k()

    if Suzuki_peak is not None:
        if Suzuki_peak < 0 or Suzuki_peak > 1:
            raise ValueError("In update_parameters, must have 0<Suzuki_peak<=1")
        A.source.Suzuki_peak = Suzuki_peak
        A.source.k_from_peak()

    # physical
    if Kappa_h is not None:
        if Kappa_h < 0:
            raise ValueError("In PhysicalParameters, Kappa_h must be positive")
        A.physical.Kappa_h = np.float64(Kappa_h)

    if Kappa_v is not None:
        if Kappa_v < 0:
            raise ValueError("In PhysicalParameters, Kappa_v must be positive")
        A.physical.Kappa_v = np.float64(Kappa_v)

    if g is not None:
        if g <= 0:
            raise ValueError("In PhysicalParameters, g must be positive")
        A.physical.g = np.float64(g)

    if mu is not None:
        if mu <= 0:
            raise ValueError("In PhysicalParameters, mu must be positive")
        A.physical.mu = np.float64(mu)

    wind_speed = met.max_wind_speed(A.source.PlumeHeight)
    ws = met.settling_speed_value(A, 0.0)[0]
    A.met = MetParameters(wind_speed, ws)

    A.model.from_params(A.solver, A.met, A.source, A.grains, A.physical)

    if start is not None:
        A.output.start = np.float64(start)
    if stop is not None:
        A.output.stop = np.float(stop)
    if step is not None:
        A.output.step = np.float(step)

    A.output.set_altitudes()
    A.output.ChebMats(A.solver.maxN, A.source.PlumeHeight)

    return A


def save_parameters(params, file="parameters.txt"):
    if os.path.exists(file):
        print(
            "WARNING: {outname} already exists and will be replaced".format(
                outname=file
            )
        )

    with open(file, "w") as f:
        f.write(
            "AshDisperse parameters {}\n\n".format(
                datetime.datetime.now().strftime("%c")
            )
        )
        # source
        f.write("Source parameters\n")
        f.write("Name = {}\n".format(params.source.name))
        f.write("Latitude = {}\n".format(params.source.latitude))
        f.write("Longitude = {}\n".format(params.source.longitude))
        f.write("UTM code = {}\n".format(params.source.utmcode))
        f.write("Radius = {}\n".format(params.source.radius))
        f.write("Plume height = {}\n".format(params.source.PlumeHeight))
        f.write("Mass eruption rate = {}\n".format(params.source.MER))
        f.write("Duration = {}\n".format(params.source.duration))
        f.write("Suzuki k = {}\n".format(params.source.Suzuki_k))
        f.write("Suzuki peak = {}\n".format(params.source.Suzuki_peak))
        f.write("\n")

        # grains
        f.write("Grain parameters\n")
        f.write("Grain classes = {}\n".format(params.grains.bins))
        for j in range(params.grains.bins):
            f.write(
                "Grain {i}:(diameter={d},density={rho},proportion={p})\n".format(
                    i=j + 1,
                    d=params.grains.diameter[j],
                    rho=params.grains.density[j],
                    p=params.grains.proportion[j],
                )
            )
        f.write("\n")

        # solver
        f.write("Solver parameters\n")
        f.write("domX = {}\n".format(params.solver.domX))
        f.write("domY = {}\n".format(params.solver.domY))
        f.write("minN_log2 = {}\n".format(params.solver.minN_log2))
        f.write("maxN_log2 = {}\n".format(params.solver.maxN_log2))
        f.write("Nx_log2 = {}\n".format(params.solver.Nx_log2))
        f.write("Nx = {}\n".format(params.solver.Nx))
        f.write("Ny_log2 = {}\n".format(params.solver.Ny_log2))
        f.write("Ny = {}\n".format(params.solver.Ny))
        f.write("epsilon = {}\n".format(params.solver.epsilon))
        f.write("\n")

        # physical
        f.write("Physical parameters\n")
        f.write("Kappa_h = {}\n".format(params.physical.Kappa_h))
        f.write("Kappa_v = {}\n".format(params.physical.Kappa_v))
        f.write("g = {}\n".format(params.physical.g))
        f.write("mu = {}\n".format(params.physical.mu))
        f.write("\n")

        # output
        f.write("Output parameters\n")
        f.write("Lower altitude = {}\n".format(params.output.start))
        f.write("Upper altitude = {}\n".format(params.output.stop))
        f.write("Altitude step = {}\n".format(params.output.step))
        f.write("\n")
    return


def load_parameters(file):
    if not os.path.exists(file):
        raise IOError("AshDisperse parameters file {} not found".format(file))

    paramslist = open(file, "r").readlines()
    pdict = dict()
    grainlist = []
    for param in paramslist:
        param = param.rstrip("\n")
        grainval = param.split(":", 2)
        if len(grainval) == 2:
            grainnum = grainval[0].split("Grain")[1].strip(" ")
            grainprops = grainval[1].strip("()").split(",", 3)
            graindict = dict()
            for gp in grainprops:
                (prop, val) = gp.split("=", 2)
                graindict[prop.strip(" ")] = np.float64(val)
            grainlist.append(graindict)

        keyval = param.split("=", 2)
        if len(keyval) == 2:
            key = keyval[0].strip()
            value = keyval[1].strip()
            try:
                pdict[key] = np.float64(value)
            except:
                pdict[key] = value

    params = Parameters()
    params.solver = SolverParameters(
        domX=pdict["domX"],
        domY=pdict["domY"],
        minN_log2=pdict["minN_log2"],
        maxN_log2=pdict["maxN_log2"],
        epsilon=pdict["epsilon"],
        Nx_log2=pdict["Nx_log2"],
        Ny_log2=pdict["Ny_log2"],
    )

    params.grains = GrainParameters()
    for j in range(int(pdict["Grain classes"])):
        params.grains.add_grain(
            grainlist[j]["diameter"],
            grainlist[j]["density"],
            grainlist[j]["proportion"],
        )

    if "Name" in pdict:
        Name = pdict["Name"]
    else:
        Name = None
    if "Suzuki k" in pdict:
        Suzuki_k = pdict["Suzuki k"]
        Suzuki_peak = None
    elif "Suzuki peak" in pdict:
        Suzuki_k = None
        Suzuki_peak = pdict["Suzuki peak"]
    else:
        raise RuntimeError(
            "Suzuki parameters not set in {}; ".format(file)
            + "need either 'Suzuki k' or 'Suzuki peak' specified"
        )
    params.source = SourceParameters(
        pdict["Latitude"],
        pdict["Longitude"],
        pdict["UTM code"],
        radius=pdict["Radius"],
        PlumeHeight=pdict["Plume height"],
        MER=pdict["Mass eruption rate"],
        duration=pdict["Duration"],
        Suzuki_k=Suzuki_k,
        Suzuki_peak=Suzuki_peak,
        name=Name,
    )

    params.physical = PhysicalParameters(
        Kappa_h=pdict["Kappa_h"], Kappa_v=pdict["Kappa_v"], g=pdict["g"], mu=pdict["mu"]
    )

    params.output = OutputParameters(
        start=pdict["Lower altitude"],
        stop=pdict["Upper altitude"],
        step=pdict["Altitude step"],
    )
    params.output.set_altitudes()
    params.output.ChebMats(params.solver.maxN, params.source.PlumeHeight)

    # windSpeed = met.max_wind_speed(params.source.PlumeHeight)
    # ws = met.settling_speed(params, np.atleast_1d(0.0))[0]
    # params.met = MetParameters(windSpeed, ws)

    # params.model = ModelParameters()
    # params.model.from_params(params.solver,
    #                          params.met,
    #                          params.source,
    #                          params.grains,
    #                          params.physical)

    return params



# -*- coding: utf-8 -*-
"""velocities.py

This module defines a container for velocities used in AshDisperse,
stored as a numba jitclass.
"""

from collections import OrderedDict

import numpy as np
from numba import float64, int64
from numba.experimental import jitclass

# pylint: disable=C0103


Vel_spec = OrderedDict()
Vel_spec["N"] = int64[::1]
Vel_spec["z"] = float64[:, :, ::1]
Vel_spec["U"] = float64[:, :, ::1]
Vel_spec["V"] = float64[:, :, ::1]
Vel_spec["Ws"] = float64[:, :, :, ::1]
Vel_spec["dWs"] = float64[:, :, :, ::1]


@jitclass(Vel_spec)
class VelocityContainer:
    """A container for velocities at chebyshev collocation points.

    VelocityContainer contains wind and settling velocities evaluated at
    chebyshev collocation points for different degrees of approximation.

    Attributes:
        N (int[:]): Array of the degrees of the chebyshev approximations.
        z (float[:,:,:]): Altitudes of the wind data (in physical space).
                          The altitudes are split into those from ground to the
                          top of the plume (contained in z[:N[k],0,k] for
                          approximation index k) and those above (contained in
                          z[:N[k],1,k] for approximation index k).
        U (float[:,:,:]): The easting component of the wind velocity.
                          The wind velocity is split into those from ground to
                          the top of the plume (contained in U[:N[k],0,k] for
                          approximation index k) and those above (contained in
                          U[:N[k],1,k] for approximation index k).
        V (float[:,:,:]): The northin component of the wind velocity.
                          Structured as U.
        Ws (float[:,:,:,:]): The settling speed.
                             Each grain class has it's own settling speed,
                             which also varies with altitude.
                             The settling is split into those from ground to
                             the top of the plume (contained in Ws[:N[k],0,k,j]
                             for approximation index k and grain class index j)
                             and those above (contained in Ws[:N[k],1,k,j]
                             for approximation index k and grain class index
                             j).
        dWs (float[:,:,:,:]): The first derivative of the settling speed with
                              z. Structured as Ws.
    """

    def __init__(self, parameters, Met, x):
        chebIts = parameters.solver.chebIts
        maxN = parameters.solver.maxN
        bins = parameters.grains.bins
        self.N = np.zeros((chebIts), dtype=np.int64)
        self.z = np.zeros((maxN, 2, chebIts), dtype=np.float64)
        self.U = -999 * np.ones((maxN, 2, chebIts), dtype=np.float64)
        self.V = -999 * np.ones((maxN, 2, chebIts), dtype=np.float64)
        self.Ws = -999 * np.ones((maxN, 2, chebIts, bins), dtype=np.float64)
        self.dWs = -999 * np.ones((maxN, 2, chebIts, bins), dtype=np.float64)

        minN_log2 = parameters.solver.minN_log2
        U_scale = parameters.met.U_scale
        Ws_scale = parameters.met.Ws_scale
        H = parameters.source.PlumeHeight
        for k in range(0, chebIts):
            N = 2 ** (k + minN_log2)
            self.N[k] = np.int64(N)
            lower_z = np.zeros((N), dtype=np.float64)
            lower_z[:N] = 0.5 * (x[:N, k] + 1) * H
            self.z[:N, 0, k] = lower_z[:N]
            self.U[:N, 0, k] = Met.wind_U_array(lower_z[:N], U_scale)
            self.V[:N, 0, k] = Met.wind_V_array(lower_z[:N], U_scale)
            self.Ws[:N, 0, k, :] = Met.settling_speed_array(
                parameters, lower_z[:N], Ws_scale
            )
            upper_z = np.zeros((N - 1), dtype=np.float64)
            upper_z[: N - 1] = 2 / (1 - x[: N - 1, k]) * H
            self.z[: N - 1, 1, k] = upper_z[: N - 1]
            self.U[: N - 1, 1, k] = Met.wind_U_array(upper_z[: N - 1], U_scale)
            self.V[: N - 1, 1, k] = Met.wind_V_array(upper_z[: N - 1], U_scale)
            self.Ws[: N - 1, 1, k, :] = Met.settling_speed_array(
                parameters, upper_z[: N - 1], Ws_scale
            )
            self.fd_Ws(k, H)

    def fd_Ws(self, k, H):
        """Compute derivate of settling speed.

        The first derivative dWs/dz is computed using centred difference, and
        assigned to the appropriate slice of the class attribute dWs.

        Args:
            k (int): the approximation index.
            H (float): the dimensional plume height (in metres).
        """
        N = self.N[k]
        nG = self.Ws.shape[-1]
        z_l = self.z[:N, 0, k] / H
        W_l = self.Ws[:N, 0, k, :]

        z_u = self.z[: N - 1, 1, k] / H
        W_u = self.Ws[: N - 1, 1, k, :]

        dW_l = np.empty((N, nG), dtype=np.float64)
        dW_u = np.empty((N - 1, nG), dtype=np.float64)

        for ig in range(nG):
            dW_l[0, ig] = (W_l[1, ig] - W_l[0, ig]) / (z_l[1] - z_l[0])
            dW_l[1:-1, ig] = (W_l[2:, ig] - W_l[:-2, ig]) / (z_l[2:] - z_l[:-2])
            dW_l[-1, ig] = (W_l[-1, ig] - W_l[-2, ig]) / (z_l[-1] - z_l[-2])

            dW_u[0, ig] = (W_u[1, ig] - W_u[0, ig]) / (z_u[1] - z_u[0])
            dW_u[1:-1, ig] = (W_u[2:, ig] - W_u[:-2, ig]) / (z_u[2:] - z_u[:-2])
            dW_u[-1, ig] = (W_u[-1, ig] - W_u[-2, ig]) / (z_u[-1] - z_u[-2])

        self.dWs[:N, 0, k, :] = dW_l
        self.dWs[: N - 1, 1, k, :] = dW_u


# pylint: disable=E1101
VelocityContainer_type = VelocityContainer.class_type.instance_type
